# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from .id import ID as ID
from .row import Row as Row
from .custom import *
from .vector import Vector as Vector
from .id_param import IDParam as IDParam
from .language import Language as Language
from .row_param import RowParam as RowParam
from .tokenizer import Tokenizer as Tokenizer
from .vector_param import VectorParam as VectorParam
from .columns_param import ColumnsParam as ColumnsParam
from .query_billing import QueryBilling as QueryBilling
from .write_billing import WriteBilling as WriteBilling
from .attribute_type import AttributeType as AttributeType
from .distance_metric import DistanceMetric as DistanceMetric
from .vector_encoding import VectorEncoding as VectorEncoding
from .full_text_search import FullTextSearch as FullTextSearch
from .namespace_summary import NamespaceSummary as NamespaceSummary
from .query_performance import QueryPerformance as QueryPerformance
from .attribute_schema_param import AttributeSchemaParam as AttributeSchemaParam
from .full_text_search_param import FullTextSearchParam as FullTextSearchParam
from .namespace_query_params import NamespaceQueryParams as NamespaceQueryParams
from .namespace_write_params import NamespaceWriteParams as NamespaceWriteParams
from .attribute_schema_config import AttributeSchemaConfig as AttributeSchemaConfig
from .full_text_search_config import FullTextSearchConfig as FullTextSearchConfig
from .namespace_recall_params import NamespaceRecallParams as NamespaceRecallParams
from .client_namespaces_params import ClientNamespacesParams as ClientNamespacesParams
from .include_attributes_param import IncludeAttributesParam as IncludeAttributesParam
from .namespace_query_response import NamespaceQueryResponse as NamespaceQueryResponse
from .namespace_write_response import NamespaceWriteResponse as NamespaceWriteResponse
from .namespace_recall_response import NamespaceRecallResponse as NamespaceRecallResponse
from .namespace_schema_response import NamespaceSchemaResponse as NamespaceSchemaResponse
from .namespace_multi_query_params import NamespaceMultiQueryParams as NamespaceMultiQueryParams
from .attribute_schema_config_param import AttributeSchemaConfigParam as AttributeSchemaConfigParam
from .full_text_search_config_param import FullTextSearchConfigParam as FullTextSearchConfigParam
from .namespace_delete_all_response import NamespaceDeleteAllResponse as NamespaceDeleteAllResponse
from .namespace_multi_query_response import NamespaceMultiQueryResponse as NamespaceMultiQueryResponse
from .namespace_update_schema_params import NamespaceUpdateSchemaParams as NamespaceUpdateSchemaParams
from .namespace_update_schema_response import NamespaceUpdateSchemaResponse as NamespaceUpdateSchemaResponse
from .namespace_hint_cache_warm_response import NamespaceHintCacheWarmResponse as NamespaceHintCacheWarmResponse



from turbopuffer import Turbopuffer
from tests.custom import test_prefix


def test_patches(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "patches")

    ns.write(
        upsert_rows=[
            {"id": 1, "vector": [1, 1]},
            {"id": 2, "vector": [2, 2]},
        ],
        distance_metric="euclidean_squared",
    )

    ns.write(
        patch_rows=[
            {"id": 1, "a": 1},
            {"id": 2, "b": 2},
        ],
    )

    ns.write(
        patch_rows=[
            {"id": 1, "b": 1},
            {"id": 2, "a": 2},
        ],
    )

    result = ns.query(rank_by=("id", "asc"), include_attributes=["a", "b"], top_k=10)
    assert result.rows is not None
    assert len(result.rows) == 2
    print(result.rows)
    assert result.rows[0].id == 1
    assert result.rows[0].a == 1
    assert result.rows[0].b == 1
    assert result.rows[1].id == 2
    assert result.rows[1].a == 2
    assert result.rows[1].b == 2

    ns.write(
        patch_columns={
            "id": [1, 2],
            "a": [11, 22],
            "c": [1, 2],
        },
    )

    result = ns.query(rank_by=("id", "asc"), include_attributes=["a", "b", "c"], top_k=10)
    assert result.rows is not None
    assert len(result.rows) == 2
    assert result.rows[0].id == 1
    assert result.rows[0].a == 11
    assert result.rows[0].b == 1
    assert result.rows[0].c == 1
    assert result.rows[1].id == 2
    assert result.rows[1].a == 22
    assert result.rows[1].b == 2
    assert result.rows[1].c == 2



import pytest

import turbopuffer
from turbopuffer import Turbopuffer
from tests.custom import test_prefix
from turbopuffer.types import AttributeSchemaConfigParam


def test_schema(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "schema")

    # Upsert some data
    ns.write(
        upsert_rows=[{"id": 2, "vector": [2, 2]}, {"id": 7, "vector": [0.7, 0.7], "hello": "world", "test": "rows"}],
        distance_metric="euclidean_squared",
    )

    # Get the schema for the namespace
    schema = ns.schema()
    for attr in ["hello", "test"]:
        assert schema[attr].type == "string"
        assert schema[attr].filterable
        assert schema[attr].full_text_search is None

    # Write an update to the schema making 'hello' not filterable
    updated_hello_schema = AttributeSchemaConfigParam(type="string", filterable=False)
    updated_schema = ns.update_schema(schema={"hello": updated_hello_schema})
    assert not updated_schema["hello"].filterable

    # If we try to query using a filter on 'hello', we should get an error
    # since it's not filterable anymore
    with pytest.raises(turbopuffer.TurbopufferError):
        ns.query(rank_by=("vector", "ANN", [2, 2]), top_k=10, filters=("hello", "Eq", "foobar"))



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal, TypeAlias

__all__ = ["Language"]

Language: TypeAlias = Literal[
    "arabic",
    "danish",
    "dutch",
    "english",
    "finnish",
    "french",
    "german",
    "greek",
    "hungarian",
    "italian",
    "norwegian",
    "portuguese",
    "romanian",
    "russian",
    "spanish",
    "swedish",
    "tamil",
    "turkish",
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal, TypeAlias

__all__ = ["VectorEncoding"]

VectorEncoding: TypeAlias = Literal["float", "base64"]



from turbopuffer import Turbopuffer
from tests.custom import test_prefix


def test_copy_from_namespace(tpuf: Turbopuffer):
    ns1_name = test_prefix + "copy_from_namespace_1"
    ns1 = tpuf.namespace(ns1_name)
    ns2 = tpuf.namespace(test_prefix + "copy_from_namespace_2")

    ns1.write(
        upsert_columns={
            "id": [1, 2, 3, 4],
            "vector": [[0.1, 0.1], [0.2, 0.2], [0.3, 0.3], [0.4, 0.4]],
            "fact_id": ["a", "b", "c", "d"],
        },
        distance_metric="euclidean_squared",
    )

    ns2.write(copy_from_namespace=ns1_name)

    query_results = ns2.query(rank_by=("vector", "ANN", [0.1, 0.1]), top_k=10)
    assert query_results.rows is not None
    assert len(query_results.rows) == 4



from typing import Any

try:
    import orjson

    loads = orjson.loads

    def dumps(obj: Any) -> bytes:
        return orjson.dumps(obj)

except ImportError:
    import json

    loads = json.loads

    def dumps(obj: Any) -> bytes:
        return json.dumps(obj).encode()



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List, Optional

from .row import Row
from .._models import BaseModel
from .query_billing import QueryBilling
from .query_performance import QueryPerformance

__all__ = ["NamespaceMultiQueryResponse", "Result"]


class Result(BaseModel):
    aggregations: Optional[Dict[str, object]] = None

    rows: Optional[List[Row]] = None


class NamespaceMultiQueryResponse(BaseModel):
    billing: QueryBilling
    """The billing information for a query."""

    performance: QueryPerformance
    """The performance information for a query."""

    results: List[Result]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from ..lib.performance import ClientPerformance

__all__ = ["QueryPerformance"]


class QueryPerformance(ClientPerformance):
    approx_namespace_size: int
    """the approximate number of documents in the namespace."""

    cache_hit_ratio: float
    """The ratio of cache hits to total cache lookups."""

    cache_temperature: str
    """A qualitative description of the cache hit ratio (`hot`, `warm`, or `cold`)."""

    exhaustive_search_count: int
    """The number of unindexed documents processed by the query."""

    query_execution_ms: int
    """
    Request time measured on the server, excluding time spent waiting due to the
    namespace concurrency limit.
    """

    server_total_ms: int
    """
    Request time measured on the server, including time spent waiting for other
    queries to complete if the namespace was at its concurrency limit.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import TypedDict

from .language import Language
from .tokenizer import Tokenizer

__all__ = ["FullTextSearchConfigParam"]


class FullTextSearchConfigParam(TypedDict, total=False):
    b: float
    """The `b` document length normalization parameter for BM25. Defaults to `0.75`."""

    case_sensitive: bool
    """Whether searching is case-sensitive.

    Defaults to `false` (i.e. case-insensitive).
    """

    k1: float
    """The `k1` term saturation parameter for BM25. Defaults to `1.2`."""

    language: Language
    """Describes the language of a text attribute. Defaults to `english`."""

    remove_stopwords: bool
    """Removes common words from the text based on language.

    Defaults to `true` (i.e. remove common words).
    """

    stemming: bool
    """Language-specific stemming for the text.

    Defaults to `false` (i.e., do not stem).
    """

    tokenizer: Tokenizer
    """The tokenizer to use for full-text search on an attribute."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict
from typing_extensions import TypedDict

from .attribute_schema_param import AttributeSchemaParam

__all__ = ["NamespaceUpdateSchemaParams"]


class NamespaceUpdateSchemaParams(TypedDict, total=False):
    namespace: str

    schema: Dict[str, AttributeSchemaParam]
    """The desired schema for the namespace."""



import uuid
from typing import List, Union

import numpy
import pytest

import turbopuffer
from turbopuffer import NOT_GIVEN, NotGiven, Turbopuffer
from tests.custom import test_prefix
from turbopuffer.types import (
    Row,
    Vector,
    RowParam,
    ColumnsParam,
    QueryBilling,
    VectorEncoding,
    NamespaceQueryResponse,
    namespace_query_params,
)
from turbopuffer.lib.vector import b64decode_vector, b64encode_vector
from turbopuffer._utils._transform import transform, async_transform


@pytest.mark.xdist_group(name="group1")
def test_upsert_rows(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "client_test")
    assert str(ns) == f"tpuf-namespace:{test_prefix}client_test"

    # Test upsert multiple dict rows
    ns.write(
        upsert_rows=[
            {"id": 2, "vector": [2, 2]},
            {"id": 7, "vector": [0.7, 0.7], "hello": "world"},
        ],
        distance_metric="euclidean_squared",
    )

    # Test upsert lazy row iterator
    ns.write(upsert_rows=[{"id": i, "vector": [i / 10, i / 10], "test": "rows"} for i in range(10, 100)])

    # Test upsert single rows should fail
    try:
        ns.write(
            upsert_rows={"id": 2, "vector": [2, 2]}  # type: ignore
        )
        raise AssertionError("Upserting single row should not be allowed")
    except turbopuffer.UnprocessableEntityError as err:
        assert "invalid type: map, expected a sequence" in err.message
    try:
        ns.write(
            upsert_rows={"id": 7, "vector": [0.7, 0.7], "hello": "world", "test": "rows"}  # type: ignore
        )
        raise AssertionError("Upserting single row should not be allowed")
    except turbopuffer.UnprocessableEntityError as err:
        assert "invalid type: map, expected a sequence" in err.message

    # Check to make sure the vectors were stored as expected
    results = ns.query(top_k=100, rank_by=("id", "asc"), include_attributes=True)
    assert results.rows is not None
    assert len(results.rows) == 92, "Got wrong number of vectors back"
    assert dict(results.rows[0]) == {"id": 2, "vector": [2.0, 2.0]}
    assert dict(results.rows[1]) == {"id": 7, "vector": [0.7, 0.7], "hello": "world"}
    for i in range(10, 100):
        assert dict(results.rows[i - 8]) == {"id": i, "vector": [i / 10, i / 10], "test": "rows"}

    # Check to ensure recall is working
    ns.recall()


@pytest.mark.xdist_group(name="group1")
def test_delete_vectors(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "client_test")

    # Test delete
    ns.write(deletes=[2])
    results = ns.query(top_k=100, rank_by=("id", "asc"), include_attributes=True)
    assert results.rows is not None
    assert len(results.rows) == 91, "Got wrong number of vectors back"
    assert dict(results.rows[0]) == {"id": 7, "vector": [0.7, 0.7], "hello": "world"}

    ns.write(deletes=[10])
    results = ns.query(top_k=100, rank_by=("id", "asc"), include_attributes=True)
    assert results.rows is not None
    assert len(results.rows) == 90, "Got wrong number of vectors back"
    assert dict(results.rows[1]) == {"id": 11, "vector": [1.1, 1.1], "test": "rows"}

    ns.write(deletes=[11])
    results = ns.query(top_k=100, rank_by=("id", "asc"), include_attributes=True)
    assert results.rows is not None
    assert len(results.rows) == 89, "Got wrong number of vectors back"
    assert dict(results.rows[1]) == {"id": 12, "vector": [1.2, 1.2], "test": "rows"}

    ns.write(deletes=[12])
    results = ns.query(top_k=100, rank_by=("id", "asc"), include_attributes=True)
    assert results.rows is not None
    assert len(results.rows) == 88, "Got wrong number of vectors back"
    assert dict(results.rows[1]) == {"id": 13, "vector": [1.3, 1.3], "test": "rows"}

    # Test delete single row
    ns.write(deletes=[13])
    # Test delete multi row
    ns.write(deletes=[14, 15, 16])

    # Check to make sure the vectors were removed
    results = ns.query(top_k=100, rank_by=("id", "asc"), include_attributes=True)
    assert results.rows is not None
    assert len(results.rows) == 84, "Got wrong number of vectors back"
    assert dict(results.rows[0]) == {"id": 7, "vector": [0.7, 0.7], "hello": "world"}
    for i in range(17, 100):
        assert dict(results.rows[i - 16]) == {"id": i, "vector": [i / 10, i / 10], "test": "rows"}


@pytest.mark.xdist_group(name="group1")
def test_upsert_columns(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "client_test")

    ids = [0, 1, 2, 3]
    vectors = [[0.0, 0.0], [0.1, 0.1], [0.2, 0.2], [0.3, 0.3]]
    attributes = {
        "key1": ["zero", "one", "two", "three"],
        "key2": [" ", "a", "b", "c"],
        "test": ["cols", "cols", "cols", "cols"],
    }

    # Test upsert columns with named args
    ns.write(upsert_columns={"id": ids, "vector": vectors, **attributes})

    # Test upsert dict columns
    ns.write(
        upsert_columns={
            "id": [0, 1, 2, 3],
            "vector": [[0.0, 0.0], [0.1, 0.1], [0.2, 0.2], [0.3, 0.3]],
            "key1": ["zero", "one", "two", "three"],
            "key2": [" ", "a", "b", "c"],
            "test": ["cols", "cols", "cols", "cols"],
        }
    )

    # Test upsert typed columns
    ns.write(
        upsert_columns={
            "id": [4, 5, 6],
            "vector": [[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]],
            "key1": ["one", "two", "three"],
            "key2": ["a", "b", "c"],
            "test": ["cols", "cols", "cols"],
        }
    )

    ns.write(upsert_rows=[{"id": i, "vector": [i / 10, i / 10], "test": "cols"} for i in range(10, 100)])

    # Check to make sure the vectors were stored as expected
    results = ns.query(top_k=100, rank_by=("id", "asc"), include_attributes=True)
    assert results.rows is not None
    assert len(results.rows) == 98, "Got wrong number of vectors back"
    test_count = 0
    for row in results.rows:
        if hasattr(row, "test") and row.test == "cols":
            test_count += 1
            assert isinstance(row.id, int)
            if row.id >= 10:
                assert dict(row) == {"id": row.id, "vector": [row.id / 10, row.id / 10], "test": "cols"}
        else:
            assert dict(row) == {"id": 7, "vector": [0.7, 0.7], "hello": "world"}
    assert test_count == 97, "Found wrong number of test cols"


@pytest.mark.xdist_group(name="group1")
def test_query_vectors(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "client_test")

    def check_result(row: Row, expected: Row):
        assert len(row) == len(expected)
        assert row.id == expected.id
        if isinstance(expected.vector, list):
            assert isinstance(row.vector, list)
            assert abs(row.vector[0] - expected.vector[0]) < 0.000001
            assert abs(row.vector[1] - expected.vector[1]) < 0.000001
        else:
            assert row.vector == expected.vector
        if "$dist" in expected:
            assert abs(row["$dist"] - expected["$dist"]) < 0.000001
        else:
            assert "$dist" not in row

    def check_results(vector_set: NamespaceQueryResponse, expected: List[Row]):
        assert vector_set.rows is not None
        assert len(vector_set.rows) == len(expected)
        for i in range(len(vector_set.rows)):
            check_result(vector_set.rows[i], expected[i])

    expected = [
        Row.from_dict({"id": 7, "vector": [0.7, 0.7], "hello": "world", "$dist": 0.01}),
        Row.from_dict({"id": 10, "vector": [1.0, 1.0], "$dist": 0.13}),
        Row.from_dict({"id": 11, "vector": [1.1, 1.1], "$dist": 0.25}),
        Row.from_dict({"id": 3, "vector": [0.3, 0.3], "$dist": 0.41}),
        Row.from_dict({"id": 6, "vector": [0.3, 0.3], "$dist": 0.41}),
    ]

    # Test normal query
    vector_set = ns.query(
        top_k=5,
        rank_by=("vector", "ANN", [0.8, 0.7]),
        include_attributes=["hello", "vector"],
    )
    check_results(vector_set, expected)
    assert vector_set.billing == QueryBilling(
        billable_logical_bytes_queried=256000000,
        billable_logical_bytes_returned=105,
    )

    # Test query with dict
    vector_set = ns.query(
        top_k=5,
        rank_by=("vector", "ANN", [0.8, 0.7]),
        include_attributes=["hello", "vector"],
    )
    check_results(vector_set, expected)

    # Test query with all attributes
    vector_set = ns.query(
        top_k=5,
        rank_by=("vector", "ANN", [0.8, 0.7]),
        include_attributes=True,
    )
    expected = [
        Row.from_dict({"id": 7, "vector": [0.7, 0.7], "$dist": 0.01, "hello": "world"}),
        Row.from_dict({"id": 10, "vector": [1.0, 1.0], "$dist": 0.13, "test": "cols"}),
        Row.from_dict({"id": 11, "vector": [1.1, 1.1], "$dist": 0.25, "test": "cols"}),
        Row.from_dict({"id": 3, "vector": [0.3, 0.3], "$dist": 0.41, "test": "cols", "key1": "three", "key2": "c"}),
        Row.from_dict({"id": 6, "vector": [0.3, 0.3], "$dist": 0.41, "test": "cols", "key1": "three", "key2": "c"}),
    ]
    check_results(vector_set, expected)

    # Test query with typed query
    vector_set = ns.query(
        top_k=5,
        rank_by=("vector", "ANN", [1.5, 1.6]),
    )
    expected = [
        Row.from_dict({"id": 15, "$dist": 0.01}),
        Row.from_dict({"id": 16, "$dist": 0.01}),
        Row.from_dict({"id": 14, "$dist": 0.05}),
        Row.from_dict({"id": 17, "$dist": 0.05}),
        Row.from_dict({"id": 18, "$dist": 0.13}),
    ]
    check_results(vector_set, expected)

    # Test query with single filter
    expected = [
        Row.from_dict({"id": 10}),
        Row.from_dict({"id": 11}),
        Row.from_dict({"id": 12}),
    ]
    vector_set = ns.query(
        rank_by=("id", "asc"),
        top_k=3,
        include_attributes=["hello"],
        filters=("id", "In", (10, 11, 12)),
    )
    check_results(vector_set, expected)

    # Test query with no vectors
    expected = [
        Row.from_dict({"id": 10}),
        Row.from_dict({"id": 11}),
        Row.from_dict({"id": 12}),
    ]
    vector_set = ns.query(
        rank_by=("id", "asc"),
        top_k=3,
        include_attributes=["hello"],
        filters=("id", "In", (10, 11, 12)),
    )
    check_results(vector_set, expected)

    # Test query with 'eventual' consistency
    vector_set = ns.query(
        rank_by=("id", "asc"),
        top_k=3,
        consistency=namespace_query_params.Consistency(level="eventual"),
    )
    assert vector_set.rows is not None
    assert len(vector_set.rows) == 3

    # Test query with no consistency dict key should fail
    try:
        ns.query(
            rank_by=("id", "asc"),
            top_k=3,
            consistency="eventual",  # type: ignore
        )
        raise AssertionError("Using consistency without a dict that has a 'level' key should not be allowed")
    except turbopuffer.UnprocessableEntityError as err:
        assert 'invalid type: string "eventual", expected struct Consistency' in err.message

    # Test query with 'nonexistentvalue' consistency should fail
    try:
        ns.query(
            rank_by=("id", "asc"),
            top_k=3,
            consistency=namespace_query_params.Consistency(level="nonexistentvalue"),  # type: ignore
        )
        raise AssertionError("Using consistency level other than 'strong' or 'eventual' should not be allowed")
    except turbopuffer.UnprocessableEntityError as err:
        assert "unknown variant `nonexistentvalue`, expected `strong` or `eventual`" in err.message


@pytest.mark.xdist_group(name="group1")
def test_delete_all(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "client_test")

    assert ns.exists() is True

    ns.delete_all()

    with pytest.raises(turbopuffer.NotFoundError):
        ns.schema()


def test_string_ids(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "string_ids")

    vec_ids = [str(uuid.uuid1()) for _ in range(0, 8)]
    key1 = ["zero", "one", "two", "three"]
    key2 = [" ", "a", "b", "c"]

    # Test upsert columns
    ns.write(
        upsert_columns={
            "id": vec_ids[0:4],
            "vector": [[0.0, 0.0], [0.1, 0.1], [0.2, 0.2], [0.3, 0.3]],
            "key1": key1,
            "key2": key2,
            "test": ["cols", "cols", "cols", "cols"],
        },
        distance_metric="euclidean_squared",
    )

    # Test upsert rows
    ns.write(
        upsert_rows=[
            {"id": id, "vector": [(i + 4) / 10, (i + 4) / 10], "test": "rows", "hello": "world"}
            for i, id in enumerate(vec_ids[4:])
        ],
        distance_metric="euclidean_squared",
    )

    # Check to make sure the vectors were stored as expected
    results = ns.query(top_k=100, rank_by=("id", "asc"), include_attributes=True)
    assert results.rows is not None
    assert len(results.rows) == 8, "Got wrong number of vectors back"
    for i, row in enumerate(results.rows):
        if row.test == "cols":
            assert dict(row) == {
                "id": vec_ids[i],
                "vector": [i / 10, i / 10],
                "key1": key1[i],
                "key2": key2[i],
                "test": "cols",
            }
        else:
            assert dict(row) == {"id": vec_ids[i], "vector": [i / 10, i / 10], "hello": "world", "test": "rows"}

    # Test query
    result = ns.query(top_k=5, rank_by=("vector", "ANN", [0.0, 0.0]), filters=("id", "In", vec_ids))
    expected = [
        Row.from_dict({"id": vec_ids[0], "vector": None, "$dist": 0.0}),
        Row.from_dict({"id": vec_ids[1], "vector": None, "$dist": 0.02}),
        Row.from_dict({"id": vec_ids[2], "vector": None, "$dist": 0.08}),
        Row.from_dict({"id": vec_ids[3], "vector": None, "$dist": 0.18}),
        Row.from_dict({"id": vec_ids[4], "vector": None, "$dist": 0.32}),
    ]
    assert result.rows is not None
    assert len(result.rows) == len(expected)
    for i, row in enumerate(result.rows):
        assert len(dict(row)) == 3
        assert row["id"] == expected[i]["id"]
        assert row["vector"] == expected[i]["vector"]
        assert abs(row["$dist"] - expected[i]["$dist"]) < 0.000001

    # Test delete by string id
    ns.write(deletes=[vec_ids[2]])

    # Test query 2
    result = ns.query(top_k=5, rank_by=("vector", "ANN", [0.0, 0.0]), filters=("id", "In", vec_ids))
    expected = [
        Row.from_dict({"id": vec_ids[0], "vector": None, "$dist": 0.0}),
        Row.from_dict({"id": vec_ids[1], "vector": None, "$dist": 0.02}),
        Row.from_dict({"id": vec_ids[3], "vector": None, "$dist": 0.18}),
        Row.from_dict({"id": vec_ids[4], "vector": None, "$dist": 0.32}),
        Row.from_dict({"id": vec_ids[5], "vector": None, "$dist": 0.5}),
    ]
    assert result.rows is not None
    assert len(result.rows) == len(expected)
    for i, row in enumerate(result.rows):
        assert len(row) == 3
        assert row["id"] == expected[i]["id"]
        assert row["vector"] == expected[i]["vector"]
        assert abs(row["$dist"] - expected[i]["$dist"]) < 0.000001

    ns.delete_all()


def test_attribute_types(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "client_test_types")

    # Test upsert multiple dict rows
    ns.write(
        upsert_rows=[
            {
                "id": 7,
                "vector": [0.7, 0.7],
                "count": 1,
                "users": ["jan", "simon"],
            },
            {
                "id": 9,
                "vector": [0.7, 0.7],
                "count": 2,
                "users": ["bojan", "morgan", "simon"],
            },
        ],
        distance_metric="euclidean_squared",
    )

    result = ns.query(
        top_k=5,
        rank_by=("vector", "ANN", [0.0, 0.0]),
        filters=(
            "And",
            [
                ("count", "Gt", 1),
                ("users", "In", ["simon"]),
            ],
        ),
    )
    assert result.rows is not None
    assert len(result.rows) == 1

    result = ns.query(
        top_k=5,
        rank_by=("vector", "ANN", [0.0, 0.0]),
        filters=(
            "Or",
            [
                ("count", "Eq", 1),
                ("users", "In", "bojan"),
            ],
        ),
    )
    assert result.rows is not None
    assert len(result.rows) == 2

    result = ns.query(
        top_k=5,
        rank_by=("vector", "ANN", [0.0, 0.0]),
        filters=(
            "And",
            [
                ("count", "Eq", 1),
                ("users", "In", "bojan"),
            ],
        ),
    )
    assert result.rows is not None
    assert len(result.rows) == 0

    result = ns.query(
        top_k=5,
        rank_by=("vector", "ANN", [0.0, 0.0]),
        filters=(
            "And",
            [
                ("count", "Eq", 1),
                (
                    "Or",
                    [
                        ("users", "In", "bojan"),
                        ("count", "Gt", 0),
                    ],
                ),
            ],
        ),
    )
    assert result.rows is not None
    assert len(result.rows) == 1


def test_not_found_error(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "not_found")

    with pytest.raises(turbopuffer.NotFoundError):
        ns.query(
            top_k=5,
            rank_by=("vector", "ANN", [0.0, 0.0]),
        )


def test_list_in_empty_namespace(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "client_test")
    assert str(ns) == f"tpuf-namespace:{test_prefix}client_test"

    # Test upsert multiple dict rows
    ns.write(
        upsert_rows=[
            {"id": 2, "vector": [2, 2]},
            {"id": 7, "vector": [1, 1]},
        ],
        distance_metric="euclidean_squared",
    )

    ns.write(deletes=[2, 7])

    result = ns.query(top_k=100, rank_by=("id", "asc"))
    assert result.rows is not None
    assert len(result.rows) == 0


def test_delete_by_filter(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "delete_by_filter")

    ns.write(
        upsert_rows=[
            {"id": 5, "vector": [0.7, 0.7], "a": 0},
            {"id": 6, "vector": [0.7, 0.7], "a": 1},
            {"id": 7, "vector": [0.7, 0.7], "a": 2},
            {"id": 8, "vector": [0.7, 0.7], "a": 1},
            {"id": 9, "vector": [0.7, 0.7], "a": 3},
        ],
        distance_metric="euclidean_squared",
    )

    result = ns.write(
        delete_by_filter=("a", "Eq", 42)  # no-op, nothing matches
    )
    assert result.rows_affected == 0

    result = ns.query(top_k=100, rank_by=("id", "asc"))
    assert result.rows is not None
    assert len(result.rows) == 5, "Got wrong number of vectors back"

    result = ns.write(delete_by_filter=("a", "Eq", 1))
    assert result.rows_affected == 2

    result = ns.query(top_k=100, rank_by=("id", "asc"))
    assert result.rows is not None
    assert len(result.rows) == 3, "Got wrong number of vectors back"
    assert result.rows[0].id == 5
    assert result.rows[1].id == 7
    assert result.rows[2].id == 9

    result = ns.write(delete_by_filter=("a", "Gt", 0))
    assert result.rows_affected == 2

    result = ns.query(top_k=100, rank_by=("id", "asc"))
    assert result.rows is not None
    assert len(result.rows) == 1, "Got wrong number of vectors back"
    assert result.rows[0].id == 5

    ns.delete_all()


def test_transparent_vector_encoding():
    # Ensure that our hacked in vector encoding fast path is actually firing.
    # Due to the nature of the hack, there's a high risk of a future refactoring
    # to the Stainless specification breaking the fast path.

    transformed = transform({"id": 1, "vector": [0.1, 0.2, 0.3]}, RowParam)
    assert transformed == {"id": 1, "vector": "zczMPc3MTD6amZk+"}

    transformed = transform({"id": [1, 2], "vector": [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]}, ColumnsParam)
    assert transformed == {"id": [1, 2], "vector": ["zczMPc3MTD6amZk+", "zczMPgAAAD+amRk/"]}


@pytest.mark.asyncio
async def test_transparent_vector_encoding_async():
    # Ensure that our hacked in vector encoding fast path is actually firing.
    # Due to the nature of the hack, there's a high risk of a future refactoring
    # to the Stainless specification breaking the fast path.

    transformed = await async_transform({"id": 1, "vector": [0.1, 0.2, 0.3]}, RowParam)
    assert transformed == {"id": 1, "vector": "zczMPc3MTD6amZk+"}

    transformed = await async_transform({"id": [1, 2], "vector": [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]}, ColumnsParam)
    assert transformed == {"id": [1, 2], "vector": ["zczMPc3MTD6amZk+", "zczMPgAAAD+amRk/"]}


def test_upsert_base64_vectors(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "base64_test")
    assert str(ns) == f"tpuf-namespace:{test_prefix}base64_test"

    # Test upsert multiple typed rows
    ns.write(
        upsert_rows=[
            {
                "id": 2,
                "vector": b64encode_vector([2, 2]),
            },
            {
                "id": 3,
                "vector": b64encode_vector([0.7, 0.7]),
                "hello": "world",
            },
        ],
        distance_metric="euclidean_squared",
    )

    # Test upsert columns
    ids = [4, 5]
    vectors = [b64encode_vector([0.1, 0.1]), b64encode_vector([0.2, 0.2])]
    attributes = {
        "key1": ["zero", "one"],
        "key2": [" ", "a"],
    }
    ns.write(
        upsert_columns={
            "id": ids,
            "vector": vectors,
            **attributes,
        }
    )

    # Test upsert lazy row iterator
    ns.write(
        upsert_rows=[
            {
                "id": i,
                "vector": b64encode_vector([i / 10, i / 10]),
                "test": "rows",
            }
            for i in range(10, 100)
        ]
    )

    # Check to make sure the vectors were stored as expected
    results = ns.query(top_k=100, rank_by=("id", "asc"), include_attributes=True)
    assert results.rows is not None
    assert len(results.rows) == 94, "Got wrong number of vectors back"
    assert dict(results.rows[0]) == {"id": 2, "vector": [2.0, 2.0]}
    assert dict(results.rows[1]) == {"id": 3, "vector": [0.7, 0.7], "hello": "world"}
    assert dict(results.rows[2]) == {"id": 4, "vector": [0.1, 0.1], "key1": "zero", "key2": " "}
    assert dict(results.rows[3]) == {"id": 5, "vector": [0.2, 0.2], "key1": "one", "key2": "a"}
    for i in range(10, 100):
        assert dict(results.rows[i - 6]) == {"id": i, "vector": [i / 10, i / 10], "test": "rows"}

    ns.delete_all()


def test_query_vectors_vector_encoding_format(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "vector_encoding_format")
    assert str(ns) == f"tpuf-namespace:{test_prefix}vector_encoding_format"

    ns.write(
        upsert_rows=[
            {
                "id": 1,
                "vector": b64encode_vector([0.1, 0.2, 0.3]),
            },
        ],
        distance_metric="euclidean_squared",
    )

    vector_encodings: List[Union[NotGiven, VectorEncoding]] = [NOT_GIVEN, "float", "base64"]
    for vector_encoding in vector_encodings:
        vector_set = ns.query(
            top_k=1,
            rank_by=("vector", "ANN", [0.0, 0.0, 0.0]),
            include_attributes=True,
            vector_encoding=vector_encoding,
        )
        assert vector_set.rows is not None
        assert len(vector_set.rows) == 1
        assert vector_set.rows[0].id == 1

        # Compare using float32 to avoid precision issues if the vectors had
        # been passed through 32-bit floats.
        def float32_list(vector: Union[None, Vector]):
            assert vector is not None
            if isinstance(vector, str):
                floats = b64decode_vector(vector)
            else:
                floats = vector
            return [numpy.float32(x) for x in floats]

        assert float32_list(vector_set.rows[0].vector) == float32_list([0.1, 0.2, 0.3])


def test_exists(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "exists-test")
    assert ns.exists() is False

    ns.write(
        upsert_rows=[
            {"id": 1, "vector": [0.1, 0.2, 0.3]},
        ],
        distance_metric="euclidean_squared",
    )
    assert ns.exists() is True



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Union
from typing_extensions import Required, TypeAlias, TypedDict

from .id_param import IDParam
from .vector_param import VectorParam

__all__ = ["RowParam"]


class RowParamTyped(TypedDict, total=False):
    id: Required[IDParam]
    """An identifier for a document."""

    vector: VectorParam
    """A vector embedding associated with a document."""


RowParam: TypeAlias = Union[RowParamTyped, Dict[str, object]]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict
from typing_extensions import TypeAlias

from .attribute_schema_config import AttributeSchemaConfig

__all__ = ["NamespaceSchemaResponse"]

NamespaceSchemaResponse: TypeAlias = Dict[str, AttributeSchemaConfig]



import turbopuffer
from turbopuffer import Turbopuffer
from tests.custom import test_prefix


def test_no_cmek(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "no_cmek")

    try:
        ns.write(
            upsert_columns={
                "id": [1, 2],
                "vector": [[0.1, 0.1], [0.2, 0.2]],
            },
            distance_metric="cosine_distance",
            encryption={"cmek": {"key_name": "mykey"}},
        )
        raise AssertionError("CMEK requires additional setup")
    except turbopuffer.BadRequestError as err:
        assert "Malformed Cloud KMS crypto key: mykey" in err.message



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Iterable
from typing_extensions import Literal, Required, TypedDict

from .distance_metric import DistanceMetric
from .vector_encoding import VectorEncoding
from .include_attributes_param import IncludeAttributesParam

__all__ = ["NamespaceMultiQueryParams", "Query", "Consistency"]


class NamespaceMultiQueryParams(TypedDict, total=False):
    namespace: str

    queries: Required[Iterable[Query]]

    consistency: Consistency
    """The consistency level for a query."""

    vector_encoding: VectorEncoding
    """The encoding to use for vectors in the response."""


class Query(TypedDict, total=False):
    aggregate_by: Dict[str, object]
    """
    Aggregations to compute over all documents in the namespace that match the
    filters.
    """

    distance_metric: DistanceMetric
    """A function used to calculate vector similarity."""

    filters: object
    """Exact filters for attributes to refine search results for.

    Think of it as a SQL WHERE clause.
    """

    include_attributes: IncludeAttributesParam
    """Whether to include attributes in the response."""

    rank_by: object
    """How to rank the documents in the namespace."""

    top_k: int
    """The number of results to return."""


class Consistency(TypedDict, total=False):
    level: Literal["strong", "eventual"]
    """The query's consistency level.

    - `strong` - Strong consistency. Requires a round-trip to object storage to
      fetch the latest writes.
    - `eventual` - Eventual consistency. Does not require a round-trip to object
      storage, but may not see the latest writes.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, Union, Mapping
from typing_extensions import Self, override

import httpx

from . import _exceptions
from ._qs import Querystring
from .types import client_namespaces_params
from ._types import (
    NOT_GIVEN,
    Body,
    Omit,
    Query,
    Headers,
    Timeout,
    NotGiven,
    Transport,
    ProxiesTypes,
    RequestOptions,
)
from ._utils import (
    is_given,
    maybe_transform,
    get_async_library,
)
from ._version import __version__
from ._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from ._streaming import Stream as Stream, AsyncStream as AsyncStream
from .pagination import SyncNamespacePage, AsyncNamespacePage
from ._exceptions import APIStatusError, TurbopufferError
from ._base_client import (
    DEFAULT_MAX_RETRIES,
    SyncAPIClient,
    AsyncAPIClient,
    AsyncPaginator,
    make_request_options,
)
from .lib.namespace import (
    Namespace,
    AsyncNamespace,
    NamespaceWithRawResponse,
    AsyncNamespaceWithRawResponse,
    NamespaceWithStreamingResponse,
    AsyncNamespaceWithStreamingResponse,
)
from .types.namespace_summary import NamespaceSummary

__all__ = [
    "Timeout",
    "Transport",
    "ProxiesTypes",
    "RequestOptions",
    "Turbopuffer",
    "AsyncTurbopuffer",
    "Client",
    "AsyncClient",
]


class Turbopuffer(SyncAPIClient):
    with_raw_response: TurbopufferWithRawResponse
    with_streaming_response: TurbopufferWithStreamedResponse

    # client options
    api_key: str
    region: str
    default_namespace: str | None

    def __init__(
        self,
        *,
        api_key: str | None = None,
        region: str | None = None,
        default_namespace: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: Union[float, Timeout, None, NotGiven] = NOT_GIVEN,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        # Configure a custom httpx client.
        # We provide a `DefaultHttpxClient` class that you can pass to retain the default values we use for `limits`, `timeout` & `follow_redirects`.
        # See the [httpx documentation](https://www.python-httpx.org/api/#client) for more details.
        http_client: httpx.Client | None = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None:
        """Construct a new synchronous Turbopuffer client instance.

        This automatically infers the following arguments from their corresponding environment variables if they are not provided:
        - `api_key` from `TURBOPUFFER_API_KEY`
        - `region` from `TURBOPUFFER_REGION`
        """
        if api_key is None:
            api_key = os.environ.get("TURBOPUFFER_API_KEY")
        if api_key is None:
            raise TurbopufferError(
                "The api_key client option must be set either by passing api_key to the client or by setting the TURBOPUFFER_API_KEY environment variable"
            )
        self.api_key = api_key

        if region is None:
            region = os.environ.get("TURBOPUFFER_REGION")
        if region is None:
            raise TurbopufferError(
                "The region client option must be set either by passing region to the client or by setting the TURBOPUFFER_REGION environment variable"
            )
        self.region = region

        self.default_namespace = default_namespace

        if base_url is None:
            base_url = os.environ.get("TURBOPUFFER_BASE_URL")
        if base_url is None:
            base_url = f"https://{region}.turbopuffer.com"

        super().__init__(
            version=__version__,
            base_url=base_url,
            max_retries=max_retries,
            timeout=timeout,
            http_client=http_client,
            custom_headers=default_headers,
            custom_query=default_query,
            _strict_response_validation=_strict_response_validation,
        )

        self.with_raw_response = TurbopufferWithRawResponse(self)
        self.with_streaming_response = TurbopufferWithStreamedResponse(self)

    def namespace(self, namespace: str) -> Namespace:
        """Create a namespace resource."""
        return Namespace(self.with_options(default_namespace=namespace))

    @property
    @override
    def qs(self) -> Querystring:
        return Querystring(array_format="comma")

    @property
    @override
    def auth_headers(self) -> dict[str, str]:
        api_key = self.api_key
        return {"Authorization": f"Bearer {api_key}"}

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:
        return {
            **super().default_headers,
            "X-Stainless-Async": "false",
            **self._custom_headers,
        }

    def copy(
        self,
        *,
        api_key: str | None = None,
        region: str | None = None,
        default_namespace: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = NOT_GIVEN,
        http_client: httpx.Client | None = None,
        max_retries: int | NotGiven = NOT_GIVEN,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
        _extra_kwargs: Mapping[str, Any] = {},
    ) -> Self:
        """
        Create a new client instance re-using the same options given to the current client with optional overriding.
        """
        if default_headers is not None and set_default_headers is not None:
            raise ValueError("The `default_headers` and `set_default_headers` arguments are mutually exclusive")

        if default_query is not None and set_default_query is not None:
            raise ValueError("The `default_query` and `set_default_query` arguments are mutually exclusive")

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers

        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        http_client = http_client or self._client
        return self.__class__(
            api_key=api_key or self.api_key,
            region=region or self.region,
            default_namespace=default_namespace or self.default_namespace,
            base_url=base_url or self.base_url,
            timeout=self.timeout if isinstance(timeout, NotGiven) else timeout,
            http_client=http_client,
            max_retries=max_retries if is_given(max_retries) else self.max_retries,
            default_headers=headers,
            default_query=params,
            **_extra_kwargs,
        )

    # Alias for `copy` for nicer inline usage, e.g.
    # client.with_options(timeout=10).foo.create(...)
    with_options = copy

    def namespaces(
        self,
        *,
        cursor: str | NotGiven = NOT_GIVEN,
        page_size: int | NotGiven = NOT_GIVEN,
        prefix: str | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> SyncNamespacePage[NamespaceSummary]:
        """
        List namespaces.

        Args:
          cursor: Retrieve the next page of results.

          page_size: Limit the number of results per page.

          prefix: Retrieve only the namespaces that match the prefix.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self.get_api_list(
            "/v1/namespaces",
            page=SyncNamespacePage[NamespaceSummary],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "cursor": cursor,
                        "page_size": page_size,
                        "prefix": prefix,
                    },
                    client_namespaces_params.ClientNamespacesParams,
                ),
            ),
            model=NamespaceSummary,
        )

    def _get_default_namespace_path_param(self) -> str:
        from_client = self.default_namespace
        if from_client is not None:
            return from_client

        raise ValueError(
            "Missing default_namespace argument; Please provide it at the client level, e.g. Turbopuffer(default_namespace='abcd') or per method."
        )

    @override
    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> APIStatusError:
        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=body)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(err_msg, response=response, body=body)

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(err_msg, response=response, body=body)

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=body)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=body)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(err_msg, response=response, body=body)

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=body)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(err_msg, response=response, body=body)
        return APIStatusError(err_msg, response=response, body=body)


class AsyncTurbopuffer(AsyncAPIClient):
    with_raw_response: AsyncTurbopufferWithRawResponse
    with_streaming_response: AsyncTurbopufferWithStreamedResponse

    # client options
    api_key: str
    region: str
    default_namespace: str | None

    def __init__(
        self,
        *,
        api_key: str | None = None,
        region: str | None = None,
        default_namespace: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: Union[float, Timeout, None, NotGiven] = NOT_GIVEN,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        # Configure a custom httpx client.
        # We provide a `DefaultAsyncHttpxClient` class that you can pass to retain the default values we use for `limits`, `timeout` & `follow_redirects`.
        # See the [httpx documentation](https://www.python-httpx.org/api/#asyncclient) for more details.
        http_client: httpx.AsyncClient | None = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None:
        """Construct a new async AsyncTurbopuffer client instance.

        This automatically infers the following arguments from their corresponding environment variables if they are not provided:
        - `api_key` from `TURBOPUFFER_API_KEY`
        - `region` from `TURBOPUFFER_REGION`
        """
        if api_key is None:
            api_key = os.environ.get("TURBOPUFFER_API_KEY")
        if api_key is None:
            raise TurbopufferError(
                "The api_key client option must be set either by passing api_key to the client or by setting the TURBOPUFFER_API_KEY environment variable"
            )
        self.api_key = api_key

        if region is None:
            region = os.environ.get("TURBOPUFFER_REGION")
        if region is None:
            raise TurbopufferError(
                "The region client option must be set either by passing region to the client or by setting the TURBOPUFFER_REGION environment variable"
            )
        self.region = region

        self.default_namespace = default_namespace

        if base_url is None:
            base_url = os.environ.get("TURBOPUFFER_BASE_URL")
        if base_url is None:
            base_url = f"https://{region}.turbopuffer.com"

        super().__init__(
            version=__version__,
            base_url=base_url,
            max_retries=max_retries,
            timeout=timeout,
            http_client=http_client,
            custom_headers=default_headers,
            custom_query=default_query,
            _strict_response_validation=_strict_response_validation,
        )

        self.with_raw_response = AsyncTurbopufferWithRawResponse(self)
        self.with_streaming_response = AsyncTurbopufferWithStreamedResponse(self)

    def namespace(self, namespace: str) -> AsyncNamespace:
        """Create a namespace resource."""
        return AsyncNamespace(self.with_options(default_namespace=namespace))

    @property
    @override
    def qs(self) -> Querystring:
        return Querystring(array_format="comma")

    @property
    @override
    def auth_headers(self) -> dict[str, str]:
        api_key = self.api_key
        return {"Authorization": f"Bearer {api_key}"}

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:
        return {
            **super().default_headers,
            "X-Stainless-Async": f"async:{get_async_library()}",
            **self._custom_headers,
        }

    def copy(
        self,
        *,
        api_key: str | None = None,
        region: str | None = None,
        default_namespace: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = NOT_GIVEN,
        http_client: httpx.AsyncClient | None = None,
        max_retries: int | NotGiven = NOT_GIVEN,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
        _extra_kwargs: Mapping[str, Any] = {},
    ) -> Self:
        """
        Create a new client instance re-using the same options given to the current client with optional overriding.
        """
        if default_headers is not None and set_default_headers is not None:
            raise ValueError("The `default_headers` and `set_default_headers` arguments are mutually exclusive")

        if default_query is not None and set_default_query is not None:
            raise ValueError("The `default_query` and `set_default_query` arguments are mutually exclusive")

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers

        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        http_client = http_client or self._client
        return self.__class__(
            api_key=api_key or self.api_key,
            region=region or self.region,
            default_namespace=default_namespace or self.default_namespace,
            base_url=base_url or self.base_url,
            timeout=self.timeout if isinstance(timeout, NotGiven) else timeout,
            http_client=http_client,
            max_retries=max_retries if is_given(max_retries) else self.max_retries,
            default_headers=headers,
            default_query=params,
            **_extra_kwargs,
        )

    # Alias for `copy` for nicer inline usage, e.g.
    # client.with_options(timeout=10).foo.create(...)
    with_options = copy

    def namespaces(
        self,
        *,
        cursor: str | NotGiven = NOT_GIVEN,
        page_size: int | NotGiven = NOT_GIVEN,
        prefix: str | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> AsyncPaginator[NamespaceSummary, AsyncNamespacePage[NamespaceSummary]]:
        """
        List namespaces.

        Args:
          cursor: Retrieve the next page of results.

          page_size: Limit the number of results per page.

          prefix: Retrieve only the namespaces that match the prefix.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self.get_api_list(
            "/v1/namespaces",
            page=AsyncNamespacePage[NamespaceSummary],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "cursor": cursor,
                        "page_size": page_size,
                        "prefix": prefix,
                    },
                    client_namespaces_params.ClientNamespacesParams,
                ),
            ),
            model=NamespaceSummary,
        )

    def _get_default_namespace_path_param(self) -> str:
        from_client = self.default_namespace
        if from_client is not None:
            return from_client

        raise ValueError(
            "Missing default_namespace argument; Please provide it at the client level, e.g. AsyncTurbopuffer(default_namespace='abcd') or per method."
        )

    @override
    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> APIStatusError:
        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=body)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(err_msg, response=response, body=body)

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(err_msg, response=response, body=body)

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=body)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=body)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(err_msg, response=response, body=body)

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=body)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(err_msg, response=response, body=body)
        return APIStatusError(err_msg, response=response, body=body)


class TurbopufferWithRawResponse:
    def __init__(self, client: Turbopuffer) -> None:
        self._client = client
        self.namespaces = to_raw_response_wrapper(
            client.namespaces,
        )

    def namespace(self, namespace: str) -> NamespaceWithRawResponse:
        """Create a namespace resource."""
        return NamespaceWithRawResponse(self._client.with_options(default_namespace=namespace))


class AsyncTurbopufferWithRawResponse:
    def __init__(self, client: AsyncTurbopuffer) -> None:
        self._client = client
        self.namespaces = async_to_raw_response_wrapper(
            client.namespaces,
        )

    def namespace(self, namespace: str) -> AsyncNamespaceWithRawResponse:
        """Create a namespace resource."""
        return AsyncNamespaceWithRawResponse(self._client.with_options(default_namespace=namespace))


class TurbopufferWithStreamedResponse:
    def __init__(self, client: Turbopuffer) -> None:
        self._client = client
        self.namespaces = to_streamed_response_wrapper(
            client.namespaces,
        )

    def namespace(self, namespace: str) -> NamespaceWithStreamingResponse:
        """Create a namespace resource."""
        return NamespaceWithStreamingResponse(self._client.with_options(default_namespace=namespace))


class AsyncTurbopufferWithStreamedResponse:
    def __init__(self, client: AsyncTurbopuffer) -> None:
        self._client = client
        self.namespaces = async_to_streamed_response_wrapper(
            client.namespaces,
        )

    def namespace(self, namespace: str) -> AsyncNamespaceWithStreamingResponse:
        """Create a namespace resource."""
        return AsyncNamespaceWithStreamingResponse(self._client.with_options(default_namespace=namespace))


Client = Turbopuffer

AsyncClient = AsyncTurbopuffer



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from tests.utils import assert_matches_type
from turbopuffer import Turbopuffer, AsyncTurbopuffer
from turbopuffer.types import (
    NamespaceQueryResponse,
    NamespaceWriteResponse,
    NamespaceRecallResponse,
    NamespaceSchemaResponse,
    NamespaceDeleteAllResponse,
    NamespaceMultiQueryResponse,
    NamespaceUpdateSchemaResponse,
    NamespaceHintCacheWarmResponse,
)

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestNamespaces:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @pytest.mark.skip()
    @parametrize
    def test_method_delete_all(self, client: Turbopuffer) -> None:
        namespace = client.namespace("namespace").delete_all()
        assert_matches_type(NamespaceDeleteAllResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_raw_response_delete_all(self, client: Turbopuffer) -> None:
        response = client.namespace("namespace").with_raw_response.delete_all()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        namespace = response.parse()
        assert_matches_type(NamespaceDeleteAllResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_streaming_response_delete_all(self, client: Turbopuffer) -> None:
        with client.namespace("namespace").with_streaming_response.delete_all() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            namespace = response.parse()
            assert_matches_type(NamespaceDeleteAllResponse, namespace, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    def test_path_params_delete_all(self, client: Turbopuffer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `namespace` but received ''"):
            client.namespace("").with_raw_response.delete_all()

    @pytest.mark.skip()
    @parametrize
    def test_method_hint_cache_warm(self, client: Turbopuffer) -> None:
        namespace = client.namespace("namespace").hint_cache_warm()
        assert_matches_type(NamespaceHintCacheWarmResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_raw_response_hint_cache_warm(self, client: Turbopuffer) -> None:
        response = client.namespace("namespace").with_raw_response.hint_cache_warm()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        namespace = response.parse()
        assert_matches_type(NamespaceHintCacheWarmResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_streaming_response_hint_cache_warm(self, client: Turbopuffer) -> None:
        with client.namespace("namespace").with_streaming_response.hint_cache_warm() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            namespace = response.parse()
            assert_matches_type(NamespaceHintCacheWarmResponse, namespace, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    def test_path_params_hint_cache_warm(self, client: Turbopuffer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `namespace` but received ''"):
            client.namespace("").with_raw_response.hint_cache_warm()

    @pytest.mark.skip()
    @parametrize
    def test_method_multi_query(self, client: Turbopuffer) -> None:
        namespace = client.namespace("namespace").multi_query(
            namespace="namespace",
            queries=[{}],
        )
        assert_matches_type(NamespaceMultiQueryResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_method_multi_query_with_all_params(self, client: Turbopuffer) -> None:
        namespace = client.namespace("namespace").multi_query(
            namespace="namespace",
            queries=[
                {
                    "aggregate_by": {"foo": "bar"},
                    "distance_metric": "cosine_distance",
                    "filters": {},
                    "include_attributes": True,
                    "rank_by": {},
                    "top_k": 0,
                }
            ],
            consistency={"level": "strong"},
            vector_encoding="float",
        )
        assert_matches_type(NamespaceMultiQueryResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_raw_response_multi_query(self, client: Turbopuffer) -> None:
        response = client.namespace("namespace").with_raw_response.multi_query(
            namespace="namespace",
            queries=[{}],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        namespace = response.parse()
        assert_matches_type(NamespaceMultiQueryResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_streaming_response_multi_query(self, client: Turbopuffer) -> None:
        with client.namespace("namespace").with_streaming_response.multi_query(
            namespace="namespace",
            queries=[{}],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            namespace = response.parse()
            assert_matches_type(NamespaceMultiQueryResponse, namespace, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    def test_path_params_multi_query(self, client: Turbopuffer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `namespace` but received ''"):
            client.namespace("namespace").with_raw_response.multi_query(
                namespace="",
                queries=[{}],
            )

    @pytest.mark.skip()
    @parametrize
    def test_method_query(self, client: Turbopuffer) -> None:
        namespace = client.namespace("namespace").query()
        assert_matches_type(NamespaceQueryResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_method_query_with_all_params(self, client: Turbopuffer) -> None:
        namespace = client.namespace("namespace").query(
            consistency={"level": "strong"},
            distance_metric="cosine_distance",
            include_attributes=True,
            rank_by=("id", "asc"),
            top_k=0,
            vector_encoding="float",
        )
        assert_matches_type(NamespaceQueryResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_raw_response_query(self, client: Turbopuffer) -> None:
        response = client.namespace("namespace").with_raw_response.query()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        namespace = response.parse()
        assert_matches_type(NamespaceQueryResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_streaming_response_query(self, client: Turbopuffer) -> None:
        with client.namespace("namespace").with_streaming_response.query() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            namespace = response.parse()
            assert_matches_type(NamespaceQueryResponse, namespace, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    def test_path_params_query(self, client: Turbopuffer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `namespace` but received ''"):
            client.namespace("").with_raw_response.query()

    @pytest.mark.skip()
    @parametrize
    def test_method_recall(self, client: Turbopuffer) -> None:
        namespace = client.namespace("namespace").recall()
        assert_matches_type(NamespaceRecallResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_method_recall_with_all_params(self, client: Turbopuffer) -> None:
        namespace = client.namespace("namespace").recall(
            num=0,
            queries=[0],
            top_k=0,
        )
        assert_matches_type(NamespaceRecallResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_raw_response_recall(self, client: Turbopuffer) -> None:
        response = client.namespace("namespace").with_raw_response.recall()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        namespace = response.parse()
        assert_matches_type(NamespaceRecallResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_streaming_response_recall(self, client: Turbopuffer) -> None:
        with client.namespace("namespace").with_streaming_response.recall() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            namespace = response.parse()
            assert_matches_type(NamespaceRecallResponse, namespace, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    def test_path_params_recall(self, client: Turbopuffer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `namespace` but received ''"):
            client.namespace("").with_raw_response.recall()

    @pytest.mark.skip()
    @parametrize
    def test_method_schema(self, client: Turbopuffer) -> None:
        namespace = client.namespace("namespace").schema()
        assert_matches_type(NamespaceSchemaResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_raw_response_schema(self, client: Turbopuffer) -> None:
        response = client.namespace("namespace").with_raw_response.schema()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        namespace = response.parse()
        assert_matches_type(NamespaceSchemaResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_streaming_response_schema(self, client: Turbopuffer) -> None:
        with client.namespace("namespace").with_streaming_response.schema() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            namespace = response.parse()
            assert_matches_type(NamespaceSchemaResponse, namespace, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    def test_path_params_schema(self, client: Turbopuffer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `namespace` but received ''"):
            client.namespace("").with_raw_response.schema()

    @pytest.mark.skip()
    @parametrize
    def test_method_update_schema(self, client: Turbopuffer) -> None:
        namespace = client.namespace("namespace").update_schema()
        assert_matches_type(NamespaceUpdateSchemaResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_method_update_schema_with_all_params(self, client: Turbopuffer) -> None:
        namespace = client.namespace("namespace").update_schema(
            schema={"foo": "string"},
        )
        assert_matches_type(NamespaceUpdateSchemaResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_raw_response_update_schema(self, client: Turbopuffer) -> None:
        response = client.namespace("namespace").with_raw_response.update_schema()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        namespace = response.parse()
        assert_matches_type(NamespaceUpdateSchemaResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_streaming_response_update_schema(self, client: Turbopuffer) -> None:
        with client.namespace("namespace").with_streaming_response.update_schema() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            namespace = response.parse()
            assert_matches_type(NamespaceUpdateSchemaResponse, namespace, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    def test_path_params_update_schema(self, client: Turbopuffer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `namespace` but received ''"):
            client.namespace("").with_raw_response.update_schema()

    @pytest.mark.skip()
    @parametrize
    def test_method_write(self, client: Turbopuffer) -> None:
        namespace = client.namespace("namespace").write()
        assert_matches_type(NamespaceWriteResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_method_write_with_all_params(self, client: Turbopuffer) -> None:
        namespace = client.namespace("namespace").write(
            copy_from_namespace="copy_from_namespace",
            deletes=["182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e"],
            distance_metric="cosine_distance",
            encryption={"cmek": {"key_name": "key_name"}},
            patch_columns={
                "id": ["182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e"],
                "vector": [[0]],
            },
            patch_rows=[
                {
                    "id": "182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
                    "vector": [0],
                }
            ],
            schema={"foo": "string"},
            upsert_columns={
                "id": ["182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e"],
                "vector": [[0]],
            },
            upsert_rows=[
                {
                    "id": "182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
                    "vector": [0],
                }
            ],
        )
        assert_matches_type(NamespaceWriteResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_raw_response_write(self, client: Turbopuffer) -> None:
        response = client.namespace("namespace").with_raw_response.write()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        namespace = response.parse()
        assert_matches_type(NamespaceWriteResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_streaming_response_write(self, client: Turbopuffer) -> None:
        with client.namespace("namespace").with_streaming_response.write() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            namespace = response.parse()
            assert_matches_type(NamespaceWriteResponse, namespace, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    def test_path_params_write(self, client: Turbopuffer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `namespace` but received ''"):
            client.namespace("").with_raw_response.write()


class TestAsyncNamespaces:
    parametrize = pytest.mark.parametrize("async_client", [False, True], indirect=True, ids=["loose", "strict"])

    @pytest.mark.skip()
    @parametrize
    async def test_method_delete_all(self, async_client: AsyncTurbopuffer) -> None:
        namespace = await async_client.namespace("namespace").delete_all()
        assert_matches_type(NamespaceDeleteAllResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_raw_response_delete_all(self, async_client: AsyncTurbopuffer) -> None:
        response = await async_client.namespace("namespace").with_raw_response.delete_all()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        namespace = await response.parse()
        assert_matches_type(NamespaceDeleteAllResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_streaming_response_delete_all(self, async_client: AsyncTurbopuffer) -> None:
        async with async_client.namespace("namespace").with_streaming_response.delete_all() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            namespace = await response.parse()
            assert_matches_type(NamespaceDeleteAllResponse, namespace, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    async def test_path_params_delete_all(self, async_client: AsyncTurbopuffer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `namespace` but received ''"):
            await async_client.namespace("").with_raw_response.delete_all()

    @pytest.mark.skip()
    @parametrize
    async def test_method_hint_cache_warm(self, async_client: AsyncTurbopuffer) -> None:
        namespace = await async_client.namespace("namespace").hint_cache_warm()
        assert_matches_type(NamespaceHintCacheWarmResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_raw_response_hint_cache_warm(self, async_client: AsyncTurbopuffer) -> None:
        response = await async_client.namespace("namespace").with_raw_response.hint_cache_warm()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        namespace = await response.parse()
        assert_matches_type(NamespaceHintCacheWarmResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_streaming_response_hint_cache_warm(self, async_client: AsyncTurbopuffer) -> None:
        async with async_client.namespace("namespace").with_streaming_response.hint_cache_warm() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            namespace = await response.parse()
            assert_matches_type(NamespaceHintCacheWarmResponse, namespace, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    async def test_path_params_hint_cache_warm(self, async_client: AsyncTurbopuffer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `namespace` but received ''"):
            await async_client.namespace("").with_raw_response.hint_cache_warm()

    @pytest.mark.skip()
    @parametrize
    async def test_method_multi_query(self, async_client: AsyncTurbopuffer) -> None:
        namespace = await async_client.namespace("namespace").multi_query(
            namespace="namespace",
            queries=[{}],
        )
        assert_matches_type(NamespaceMultiQueryResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_method_multi_query_with_all_params(self, async_client: AsyncTurbopuffer) -> None:
        namespace = await async_client.namespace("namespace").multi_query(
            namespace="namespace",
            queries=[
                {
                    "aggregate_by": {"foo": "bar"},
                    "distance_metric": "cosine_distance",
                    "filters": {},
                    "include_attributes": True,
                    "rank_by": {},
                    "top_k": 0,
                }
            ],
            consistency={"level": "strong"},
            vector_encoding="float",
        )
        assert_matches_type(NamespaceMultiQueryResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_raw_response_multi_query(self, async_client: AsyncTurbopuffer) -> None:
        response = await async_client.namespace("namespace").with_raw_response.multi_query(
            namespace="namespace",
            queries=[{}],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        namespace = await response.parse()
        assert_matches_type(NamespaceMultiQueryResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_streaming_response_multi_query(self, async_client: AsyncTurbopuffer) -> None:
        async with async_client.namespace("namespace").with_streaming_response.multi_query(
            namespace="namespace",
            queries=[{}],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            namespace = await response.parse()
            assert_matches_type(NamespaceMultiQueryResponse, namespace, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    async def test_path_params_multi_query(self, async_client: AsyncTurbopuffer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `namespace` but received ''"):
            await async_client.namespace("namespace").with_raw_response.multi_query(
                namespace="",
                queries=[{}],
            )

    @pytest.mark.skip()
    @parametrize
    async def test_method_query(self, async_client: AsyncTurbopuffer) -> None:
        namespace = await async_client.namespace("namespace").query()
        assert_matches_type(NamespaceQueryResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_method_query_with_all_params(self, async_client: AsyncTurbopuffer) -> None:
        namespace = await async_client.namespace("namespace").query(
            aggregate_by={"foo": "bar"},
            consistency={"level": "strong"},
            distance_metric="cosine_distance",
            filters={},
            include_attributes=True,
            rank_by={},
            top_k=0,
            vector_encoding="float",
        )
        assert_matches_type(NamespaceQueryResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_raw_response_query(self, async_client: AsyncTurbopuffer) -> None:
        response = await async_client.namespace("namespace").with_raw_response.query()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        namespace = await response.parse()
        assert_matches_type(NamespaceQueryResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_streaming_response_query(self, async_client: AsyncTurbopuffer) -> None:
        async with async_client.namespace("namespace").with_streaming_response.query() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            namespace = await response.parse()
            assert_matches_type(NamespaceQueryResponse, namespace, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    async def test_path_params_query(self, async_client: AsyncTurbopuffer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `namespace` but received ''"):
            await async_client.namespace("").with_raw_response.query()

    @pytest.mark.skip()
    @parametrize
    async def test_method_recall(self, async_client: AsyncTurbopuffer) -> None:
        namespace = await async_client.namespace("namespace").recall()
        assert_matches_type(NamespaceRecallResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_method_recall_with_all_params(self, async_client: AsyncTurbopuffer) -> None:
        namespace = await async_client.namespace("namespace").recall(
            filters={},
            num=0,
            queries=[0],
            top_k=0,
        )
        assert_matches_type(NamespaceRecallResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_raw_response_recall(self, async_client: AsyncTurbopuffer) -> None:
        response = await async_client.namespace("namespace").with_raw_response.recall()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        namespace = await response.parse()
        assert_matches_type(NamespaceRecallResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_streaming_response_recall(self, async_client: AsyncTurbopuffer) -> None:
        async with async_client.namespace("namespace").with_streaming_response.recall() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            namespace = await response.parse()
            assert_matches_type(NamespaceRecallResponse, namespace, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    async def test_path_params_recall(self, async_client: AsyncTurbopuffer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `namespace` but received ''"):
            await async_client.namespace("").with_raw_response.recall()

    @pytest.mark.skip()
    @parametrize
    async def test_method_schema(self, async_client: AsyncTurbopuffer) -> None:
        namespace = await async_client.namespace("namespace").schema()
        assert_matches_type(NamespaceSchemaResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_raw_response_schema(self, async_client: AsyncTurbopuffer) -> None:
        response = await async_client.namespace("namespace").with_raw_response.schema()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        namespace = await response.parse()
        assert_matches_type(NamespaceSchemaResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_streaming_response_schema(self, async_client: AsyncTurbopuffer) -> None:
        async with async_client.namespace("namespace").with_streaming_response.schema() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            namespace = await response.parse()
            assert_matches_type(NamespaceSchemaResponse, namespace, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    async def test_path_params_schema(self, async_client: AsyncTurbopuffer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `namespace` but received ''"):
            await async_client.namespace("").with_raw_response.schema()

    @pytest.mark.skip()
    @parametrize
    async def test_method_update_schema(self, async_client: AsyncTurbopuffer) -> None:
        namespace = await async_client.namespace("namespace").update_schema()
        assert_matches_type(NamespaceUpdateSchemaResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_method_update_schema_with_all_params(self, async_client: AsyncTurbopuffer) -> None:
        namespace = await async_client.namespace("namespace").update_schema(
            schema={"foo": "string"},
        )
        assert_matches_type(NamespaceUpdateSchemaResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_raw_response_update_schema(self, async_client: AsyncTurbopuffer) -> None:
        response = await async_client.namespace("namespace").with_raw_response.update_schema()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        namespace = await response.parse()
        assert_matches_type(NamespaceUpdateSchemaResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_streaming_response_update_schema(self, async_client: AsyncTurbopuffer) -> None:
        async with async_client.namespace("namespace").with_streaming_response.update_schema() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            namespace = await response.parse()
            assert_matches_type(NamespaceUpdateSchemaResponse, namespace, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    async def test_path_params_update_schema(self, async_client: AsyncTurbopuffer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `namespace` but received ''"):
            await async_client.namespace("").with_raw_response.update_schema()

    @pytest.mark.skip()
    @parametrize
    async def test_method_write(self, async_client: AsyncTurbopuffer) -> None:
        namespace = await async_client.namespace("namespace").write()
        assert_matches_type(NamespaceWriteResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_method_write_with_all_params(self, async_client: AsyncTurbopuffer) -> None:
        namespace = await async_client.namespace("namespace").write(
            copy_from_namespace="copy_from_namespace",
            delete_by_filter={},
            deletes=["182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e"],
            distance_metric="cosine_distance",
            encryption={"cmek": {"key_name": "key_name"}},
            patch_columns={
                "id": ["182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e"],
                "vector": [[0]],
            },
            patch_rows=[
                {
                    "id": "182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
                    "vector": [0],
                }
            ],
            schema={"foo": "string"},
            upsert_columns={
                "id": ["182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e"],
                "vector": [[0]],
            },
            upsert_rows=[
                {
                    "id": "182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
                    "vector": [0],
                }
            ],
        )
        assert_matches_type(NamespaceWriteResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_raw_response_write(self, async_client: AsyncTurbopuffer) -> None:
        response = await async_client.namespace("namespace").with_raw_response.write()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        namespace = await response.parse()
        assert_matches_type(NamespaceWriteResponse, namespace, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_streaming_response_write(self, async_client: AsyncTurbopuffer) -> None:
        async with async_client.namespace("namespace").with_streaming_response.write() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            namespace = await response.parse()
            assert_matches_type(NamespaceWriteResponse, namespace, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    async def test_path_params_write(self, async_client: AsyncTurbopuffer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `namespace` but received ''"):
            await async_client.namespace("").with_raw_response.write()



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union
from typing_extensions import TypeAlias

from .attribute_type import AttributeType
from .attribute_schema_config_param import AttributeSchemaConfigParam

__all__ = ["AttributeSchemaParam"]

AttributeSchemaParam: TypeAlias = Union[AttributeType, AttributeSchemaConfigParam]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .._models import BaseModel

__all__ = ["NamespaceRecallResponse"]


class NamespaceRecallResponse(BaseModel):
    avg_ann_count: float
    """
    The average number of documents retrieved by the approximate nearest neighbor
    searches.
    """

    avg_exhaustive_count: float
    """The average number of documents retrieved by the exhaustive searches."""

    avg_recall: float
    """The average recall of the queries."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal

from .._models import BaseModel

__all__ = ["NamespaceDeleteAllResponse"]


class NamespaceDeleteAllResponse(BaseModel):
    status: Literal["OK"]
    """The status of the request."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .._models import BaseModel

__all__ = ["QueryBilling"]


class QueryBilling(BaseModel):
    billable_logical_bytes_queried: int
    """The number of billable logical bytes queried from the namespace."""

    billable_logical_bytes_returned: int
    """The number of billable logical bytes returned from the query."""



from typing import Iterator, AsyncIterator

import pytest

from turbopuffer import Turbopuffer, AsyncTurbopuffer


@pytest.fixture(scope="session")
def tpuf() -> Iterator[Turbopuffer]:
    with Turbopuffer() as tpuf:
        yield tpuf


@pytest.fixture(scope="session")
async def async_tpuf() -> AsyncIterator[AsyncTurbopuffer]:
    yield AsyncTurbopuffer()



import pytest

import turbopuffer
from turbopuffer import Turbopuffer
from tests.custom import test_prefix
from turbopuffer.types import RankBy, AttributeSchemaParam, FullTextSearchConfigParam, AttributeSchemaConfigParam


def test_bm25(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "bm25")

    schema: dict[str, AttributeSchemaParam] = {
        "blabla": AttributeSchemaConfigParam(
            type="string",
            full_text_search=FullTextSearchConfigParam(
                language="english",
                stemming=True,
                case_sensitive=False,
                remove_stopwords=True,
            ),
        ),
    }

    # Upsert with dict
    ns.write(
        upsert_columns={
            "id": [1, 2, 3, 4],
            "vector": [[0.1, 0.1], [0.2, 0.2], [0.3, 0.3], [0.4, 0.4]],
            "blabla": [
                "the fox is quick and brown",
                "fox jumped over the lazy dog",
                "the dog is lazy and brown",
                "walrus is slow and inferior",
            ],
            "fact_id": ["a", "b", "c", "d"],
        },
        distance_metric="euclidean_squared",
        schema=schema,
    )

    # Query with dict
    result = ns.query(top_k=10, rank_by=("blabla", "BM25", "fox jumping"))
    assert result.rows is not None
    assert [item.id for item in result.rows] == [2, 1]

    # Upsert with named args
    ns.write(
        upsert_columns={
            "id": [7],
            "vector": [[0.7, 0.7]],
            "blabla": ["walruses can live up to 40 years"],
            "fact_id": ["g"],
        },
        schema=schema,
    )

    # Query with named args (and combined ranking)
    result = ns.query(
        top_k=2,
        rank_by=(
            "Sum",
            [("blabla", "BM25", "walrus tusk"), ("blabla", "BM25", "jumping fox")],
        ),
        filters=("fact_id", "NotEq", "z"),
    )
    assert result.rows is not None
    assert [item.id for item in result.rows] == [2, 4]

    # Upsert with row-based upsert format
    ns.write(
        upsert_rows=[
            {
                "id": 8,
                "vector": [0.8, 0.8],
                "blabla": "row based upsert format is cool",
                "fact_id": "h",
            },
            {
                "id": 9,
                "vector": [0.9, 0.9],
                "blabla": "dict format of row based upsert also works, but isn't typed as well",
                "fact_id": "i",
            },
        ],
        schema=schema,
    )

    # Query to make sure the new row(s) is there
    result = ns.query(
        top_k=10,
        rank_by=("blabla", "BM25", "row based upsert"),
    )
    assert result.rows is not None
    assert [item.id for item in result.rows] == [8, 9]


def test_bm25_product_operator(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "bm25_product_operator")

    try:
        ns.delete_all()
    except turbopuffer.NotFoundError:
        pass

    schema: dict[str, AttributeSchemaParam] = {
        "title": {
            "type": "string",
            "full_text_search": True,
        },
        "content": {
            "type": "string",
            "full_text_search": True,
        },
    }

    # Upsert a few docs
    ns.write(
        upsert_rows=[
            {
                "id": 1,
                "vector": [0.1, 0.1],
                "title": "the quick brown fox",
                "content": "jumped over the lazy dog",
            },
            {
                "id": 2,
                "vector": [0.2, 0.2],
                "title": "the lazy dog",
                "content": "is brown",
            },
        ],
        distance_metric="euclidean_squared",
        schema=schema,
    )

    # Try out a bunch of query variations
    queries: list[RankBy] = [
        ("Product", (0.5, ("title", "BM25", "quick brown"))),
        ("Product", (("title", "BM25", "quick brown"), 0.5)),
        (
            "Sum",
            [
                ("Product", (0.5, ("title", "BM25", "quick brown"))),
                ("content", "BM25", "brown"),
            ],
        ),
        (
            "Product",
            (
                0.5,
                (
                    "Sum",
                    [
                        ("Product", (0.5, ("title", "BM25", "quick brown"))),
                        ("content", "BM25", "brown"),
                    ],
                ),
            ),
        ),
    ]

    for query in queries:
        result = ns.query(
            rank_by=query,
            top_k=10,
        )
        assert result.rows is not None
        assert len(result.rows) > 0


def test_bm25_ContainsAllTokens(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "bm25_ContainsAllTokens")

    try:
        ns.delete_all()
    except turbopuffer.NotFoundError:
        pass

    ns.write(
        upsert_rows=[
            {
                "id": 1,
                "vector": [0.1, 0.1],
                "text": "Walruses are large marine mammals with long tusks and whiskers",
            }
        ],
        schema={
            "text": {
                "type": "string",
                "full_text_search": {
                    "stemming": True,
                },
            },
        },
        distance_metric="cosine_distance",
    )

    result = ns.query(
        rank_by=("text", "BM25", "walrus whisker"),
        filters=("text", "ContainsAllTokens", "marine mammals"),
        top_k=10,
    )
    assert result.rows is not None
    assert len(result.rows) == 1

    missing = ns.query(
        rank_by=("text", "BM25", "walrus whisker"),
        filters=("text", "ContainsAllTokens", "marine mammals short"),
        top_k=10,
    )
    assert missing.rows is not None
    assert len(missing.rows) == 0


def test_bm25_pre_tokenized_array(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "bm25_pre_tokenized_array")

    try:
        ns.delete_all()
    except turbopuffer.NotFoundError:
        pass

    schema: dict[str, AttributeSchemaParam] = {
        "content": AttributeSchemaConfigParam(
            type="[]string",
            full_text_search=FullTextSearchConfigParam(
                tokenizer="pre_tokenized_array",
            ),
        ),
    }

    ns.write(
        upsert_rows=[
            {
                "id": 1,
                "content": ["jumped", "over", "the", "lazy", "dog"],
            },
            {
                "id": 2,
                "content": ["the", "lazy", "dog", "is", "brown"],
            },
        ],
        schema=schema,
    )

    result = ns.query(
        rank_by=("content", "BM25", ["jumped"]),
        top_k=10,
    )
    assert result.rows is not None
    assert len(result.rows) == 1
    assert result.rows[0].id == 1

    result = ns.query(
        rank_by=("content", "BM25", ["dog"]),
        top_k=10,
    )
    assert result.rows is not None
    assert len(result.rows) == 2

    with pytest.raises(
        turbopuffer.APIError, match=r"""invalid input \\'jumped\\' for rank_by field "content", expecting \[\]string"""
    ):
        # Query must be an array.
        ns.query(
            rank_by=("content", "BM25", "jumped"),
            top_k=10,
        )



from turbopuffer import Turbopuffer
from tests.custom import test_prefix


def test_namespace_hint_cache_warm(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "namespace_hint_cache_warm")

    # Upsert anything
    ns.write(
        upsert_rows=[
            {"id": 2, "vector": [2, 2]},
            {"id": 7, "vector": [0.7, 0.7], "hello": "world", "test": "rows"},
        ],
        distance_metric="euclidean_squared",
    )

    result = ns.hint_cache_warm()
    assert isinstance(result.message, str)
    assert result.status in ["ACCEPTED", "OK"]



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol
from typing_extensions import override, deprecated

from ..resources import namespaces

if TYPE_CHECKING:
    from .._client import Turbopuffer, AsyncTurbopuffer


class NamespaceProtocol(Protocol):
    _client: Turbopuffer | AsyncTurbopuffer
    id: str


class NamespaceMixin:
    @property
    def id(self: NamespaceProtocol) -> str:
        assert self._client.default_namespace is not None
        return self._client.default_namespace

    @override
    def __str__(self: NamespaceProtocol) -> str:
        return f"tpuf-namespace:{self.id}"


class Namespace(namespaces.NamespacesResource, NamespaceMixin):
    def __init__(self, client: Turbopuffer) -> None:
        super().__init__(client)

    # TODO(benesch): remove these shims when sufficient time has passed (say, Sep 2025).

    @deprecated("method removed; see https://github.com/turbopuffer/turbopuffer-python/blob/main/UPGRADING.md for help")
    def metadata(self) -> None:
        raise NotImplementedError(
            "the Namespace.metadata() method has been removed; see https://github.com/turbopuffer/turbopuffer-python/blob/main/UPGRADING.md for help"
        )

    @deprecated("method removed; see https://github.com/turbopuffer/turbopuffer-python/blob/main/UPGRADING.md for help")
    def dimensions(self) -> None:
        raise NotImplementedError(
            "the Namespace.dimensions() method has been removed; see https://github.com/turbopuffer/turbopuffer-python/blob/main/UPGRADING.md for help"
        )

    @deprecated("method removed; see https://github.com/turbopuffer/turbopuffer-python/blob/main/UPGRADING.md for help")
    def approx_count(self) -> None:
        raise NotImplementedError(
            "the Namespace.approx_count() method has been removed; see https://github.com/turbopuffer/turbopuffer-python/blob/main/UPGRADING.md for help"
        )


class NamespaceWithRawResponse(namespaces.NamespacesResourceWithRawResponse, NamespaceMixin):
    def __init__(self, client: Turbopuffer) -> None:
        super().__init__(namespaces.NamespacesResource(client))


class NamespaceWithStreamingResponse(namespaces.NamespacesResourceWithStreamingResponse, NamespaceMixin):
    def __init__(self, client: Turbopuffer) -> None:
        super().__init__(namespaces.NamespacesResource(client))


class AsyncNamespace(namespaces.AsyncNamespacesResource, NamespaceMixin):
    def __init__(self, client: AsyncTurbopuffer) -> None:
        super().__init__(client)


class AsyncNamespaceWithRawResponse(namespaces.AsyncNamespacesResourceWithRawResponse, NamespaceMixin):
    def __init__(self, client: AsyncTurbopuffer) -> None:
        super().__init__(namespaces.AsyncNamespacesResource(client))


class AsyncNamespaceWithStreamingResponse(namespaces.AsyncNamespacesResourceWithStreamingResponse, NamespaceMixin):
    def __init__(self, client: AsyncTurbopuffer) -> None:
        super().__init__(namespaces.AsyncNamespacesResource(client))



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import TypedDict

from .attribute_type import AttributeType
from .full_text_search_param import FullTextSearchParam

__all__ = ["AttributeSchemaConfigParam"]


class AttributeSchemaConfigParam(TypedDict, total=False):
    ann: bool
    """Whether to create an approximate nearest neighbor index for the attribute."""

    filterable: bool
    """Whether or not the attributes can be used in filters."""

    full_text_search: FullTextSearchParam
    """Whether this attribute can be used as part of a BM25 full-text search.

    Requires the `string` or `[]string` type, and by default, BM25-enabled
    attributes are not filterable. You can override this by setting
    `filterable: true`.
    """

    type: AttributeType
    """The data type of the attribute.

    Valid values: string, int, uint, uuid, datetime, bool, []string, []int, []uint,
    []uuid, []datetime, [DIMS]f16, [DIMS]f32.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Iterable
from typing_extensions import TypedDict

__all__ = ["NamespaceRecallParams"]


class NamespaceRecallParams(TypedDict, total=False):
    namespace: str

    filters: object
    """Filter by attributes. Same syntax as the query endpoint."""

    num: int
    """The number of searches to run."""

    queries: Iterable[float]
    """Use specific query vectors for the measurement.

    If omitted, sampled from the index.
    """

    top_k: int
    """Search for `top_k` nearest neighbors."""



# Copyright © 2025, Karen Petrosyan. All rights reserved.
# Obtained from the following URL on 9 June 2025:
# https://github.com/karpetrosyan/httpx-aiohttp/tree/6bfcb1dcf13632aa5338e49b13dccf54bd51ca65
#
# The default async transport for `httpx` is quite slow, so we use this
# aiohttp-based transport instead.

import gzip
import time
import typing
import asyncio
import contextlib
from typing_extensions import override

import httpx
import aiohttp
from aiohttp import ClientTimeout
from aiohttp.client import ClientSession, ClientResponse

from .transport import MIN_GZIP_SIZE

AIOHTTP_EXC_MAP = {
    aiohttp.ServerTimeoutError: httpx.TimeoutException,
    aiohttp.ConnectionTimeoutError: httpx.ConnectTimeout,
    aiohttp.SocketTimeoutError: httpx.ReadTimeout,
    aiohttp.ClientConnectorError: httpx.ConnectError,
    aiohttp.ClientPayloadError: httpx.ReadError,
    aiohttp.ClientProxyConnectionError: httpx.ProxyError,
}


@contextlib.contextmanager
def map_aiohttp_exceptions() -> typing.Iterator[None]:
    try:
        yield
    except Exception as exc:
        mapped_exc = None

        for from_exc, to_exc in AIOHTTP_EXC_MAP.items():
            if not isinstance(exc, from_exc):  # type: ignore
                continue
            if mapped_exc is None or issubclass(to_exc, mapped_exc):  # type: ignore
                mapped_exc = to_exc

        if mapped_exc is None:  # pragma: no cover
            raise

        message = str(exc)
        raise mapped_exc(message) from exc


class AiohttpResponseStream(httpx.AsyncByteStream):
    CHUNK_SIZE = 1024 * 16

    def __init__(self, aiohttp_response: ClientResponse) -> None:
        self._aiohttp_response = aiohttp_response

    @override
    async def __aiter__(self) -> typing.AsyncIterator[bytes]:
        with map_aiohttp_exceptions():
            async for chunk in self._aiohttp_response.content.iter_chunked(self.CHUNK_SIZE):
                yield chunk

    @override
    async def aclose(self) -> None:
        with map_aiohttp_exceptions():
            await self._aiohttp_response.__aexit__(None, None, None)


class AiohttpTransport(httpx.AsyncBaseTransport):
    def __init__(self, client: typing.Union[ClientSession, typing.Callable[[], ClientSession]]) -> None:
        self.client = client

    @override
    async def handle_async_request(
        self,
        request: httpx.Request,
    ) -> httpx.Response:
        clock = request.extensions["clock"]
        timeout = request.extensions.get("timeout", {})
        sni_hostname = request.extensions.get("sni_hostname")

        if not isinstance(self.client, ClientSession):
            self.client = self.client()

        # Compress the request content if worthwhile.
        content = request.content
        clock["compress_start"] = time.monotonic()
        clock["compress_end"] = clock["compress_start"]  # overwritten later if we actually compress
        if len(request.content) > MIN_GZIP_SIZE:
            loop = asyncio.get_event_loop()
            content = await loop.run_in_executor(None, lambda: gzip.compress(content, compresslevel=1))
            request.headers["Content-Encoding"] = "gzip"
            request.headers["Content-Length"] = str(len(content))
            clock["compress_end"] = time.monotonic()

        with map_aiohttp_exceptions():
            response = await self.client.request(
                method=request.method,
                url=str(request.url),
                headers=request.headers,
                data=content,
                allow_redirects=False,
                auto_decompress=False,
                compress=False,
                timeout=ClientTimeout(
                    sock_connect=timeout.get("connect"),
                    sock_read=timeout.get("read"),
                    connect=timeout.get("pool"),
                ),
                server_hostname=sni_hostname,
            ).__aenter__()
            clock["response_headers_end"] = time.monotonic()

        httpx_response = httpx.Response(
            status_code=response.status,
            headers=response.headers,
            content=AiohttpResponseStream(response),
            request=request,
            extensions={
                # Smuggle the clock through the response object.
                "clock": clock,
            },
        )

        # Eagerly read the response content (which will be cached on the
        # response object) so that we can time the body read.
        await httpx_response.aread()
        clock["body_read_end"] = time.monotonic()

        return httpx_response

    @override
    async def aclose(self) -> None:
        if isinstance(self.client, ClientSession):
            await self.client.close()



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import List, Union
from typing_extensions import TypeAlias

__all__ = ["IncludeAttributesParam"]

IncludeAttributesParam: TypeAlias = Union[bool, List[str]]



import gzip
import time
from typing_extensions import override

import httpx

from .transport import MIN_GZIP_SIZE


class HttpxTransport(httpx.HTTPTransport):
    @override
    def handle_request(self, request: httpx.Request) -> httpx.Response:
        clock = request.extensions["clock"]

        # Compress the request content if worthwhile.
        clock["compress_start"] = time.monotonic()
        clock["compress_end"] = clock["compress_start"]  # overwritten later if we actually compress
        if len(request.content) > MIN_GZIP_SIZE:
            request._content = gzip.compress(request.content, compresslevel=1)
            request.stream = httpx.ByteStream(request._content)
            request.headers["Content-Encoding"] = "gzip"
            request.headers["Content-Length"] = str(len(request.content))
            clock["compress_end"] = time.monotonic()

        response = super().handle_request(request)
        clock["response_headers_end"] = time.monotonic()

        # Eagerly read the response content (which will be cached on the
        # response object) so that we can time the body read.
        response.read()
        clock["body_read_end"] = time.monotonic()

        # Smuggle the clock through the response object.
        response.extensions["clock"] = clock

        return response



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional

from .._models import BaseModel
from .query_billing import QueryBilling

__all__ = ["WriteBilling"]


class WriteBilling(BaseModel):
    billable_logical_bytes_written: int
    """The number of billable logical bytes written to the namespace."""

    query: Optional[QueryBilling] = None
    """The billing information for a query."""



from turbopuffer import Turbopuffer
from tests.custom import test_prefix


def test_perf_metrics(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "perf_metrics")

    ns.write(
        upsert_rows=[
            {"id": 1, "vector": [0.1, 0.1]},
            {"id": 2, "vector": [0.2, 0.2], "hello": "world", "test": "rows"},
        ],
        distance_metric="cosine_distance",
    )

    result = ns.query(
        rank_by=("vector", "ANN", [0.0, 0.3]),
        include_attributes=["hello"],
        filters=("hello", "Eq", "world"),
        top_k=10,
    )

    performance = result.performance

    assert type(performance.cache_hit_ratio) is float
    assert performance.cache_temperature in ["hot", "warm", "cold"]
    assert performance.exhaustive_search_count == 2
    assert performance.server_total_ms > 0
    assert performance.query_execution_ms > 0



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Union
from typing_extensions import TypeAlias

from .full_text_search_config import FullTextSearchConfig

__all__ = ["FullTextSearch"]

FullTextSearch: TypeAlias = Union[bool, FullTextSearchConfig]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union, Iterable
from typing_extensions import TypeAlias

__all__ = ["VectorParam"]

VectorParam: TypeAlias = Union[Iterable[float], str]



from turbopuffer import AsyncTurbopuffer
from tests.custom import test_prefix


async def test_order_by_attribute(async_tpuf: AsyncTurbopuffer):
    ns = async_tpuf.namespace(test_prefix + "order_by_attribute")

    await ns.write(
        upsert_columns={
            "id": [1, 2, 3, 4],
            "vector": [[0.1, 0.1], [0.2, 0.2], [0.3, 0.3], [0.4, 0.4]],
            "fact_id": ["a", "b", "c", "d"],
        },
        distance_metric="euclidean_squared",
    )

    results = await ns.query(rank_by=("fact_id", "asc"), top_k=10)
    assert results.rows is not None
    assert [item.id for item in results.rows] == [1, 2, 3, 4]

    results = await ns.query(rank_by=("fact_id", "desc"), top_k=10)
    assert results.rows is not None
    assert [item.id for item in results.rows] == [4, 3, 2, 1]


async def test_exists(async_tpuf: AsyncTurbopuffer):
    ns = async_tpuf.namespace(test_prefix + "exists-test-async")
    assert (await ns.exists()) is False

    await ns.write(
        upsert_rows=[
            {"id": 1, "vector": [0.1, 0.2, 0.3]},
        ],
        distance_metric="euclidean_squared",
    )
    assert (await ns.exists()) is True



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import TypedDict

__all__ = ["ClientNamespacesParams"]


class ClientNamespacesParams(TypedDict, total=False):
    cursor: str
    """Retrieve the next page of results."""

    page_size: int
    """Limit the number of results per page."""

    prefix: str
    """Retrieve only the namespaces that match the prefix."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal, TypeAlias

__all__ = ["DistanceMetric"]

DistanceMetric: TypeAlias = Literal["cosine_distance", "euclidean_squared"]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal

from .._models import BaseModel
from .write_billing import WriteBilling

__all__ = ["NamespaceWriteResponse"]


class NamespaceWriteResponse(BaseModel):
    billing: WriteBilling
    """The billing information for a write request."""

    message: str
    """A message describing the result of the write request."""

    rows_affected: int
    """The number of rows affected by the write request."""

    status: Literal["OK"]
    """The status of the request."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from tests.utils import assert_matches_type
from turbopuffer import Turbopuffer, AsyncTurbopuffer
from turbopuffer.types import NamespaceSummary
from turbopuffer.pagination import SyncNamespacePage, AsyncNamespacePage

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestClient:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @pytest.mark.skip()
    @parametrize
    def test_method_namespaces(self, client: Turbopuffer) -> None:
        client_ = client.namespaces()
        assert_matches_type(SyncNamespacePage[NamespaceSummary], client_, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_method_namespaces_with_all_params(self, client: Turbopuffer) -> None:
        client_ = client.namespaces(
            cursor="cursor",
            page_size=1,
            prefix="prefix",
        )
        assert_matches_type(SyncNamespacePage[NamespaceSummary], client_, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_raw_response_namespaces(self, client: Turbopuffer) -> None:
        response = client.with_raw_response.namespaces()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        client_ = response.parse()
        assert_matches_type(SyncNamespacePage[NamespaceSummary], client_, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_streaming_response_namespaces(self, client: Turbopuffer) -> None:
        with client.with_streaming_response.namespaces() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            client_ = response.parse()
            assert_matches_type(SyncNamespacePage[NamespaceSummary], client_, path=["response"])

        assert cast(Any, response.is_closed) is True


class TestAsyncClient:
    parametrize = pytest.mark.parametrize("async_client", [False, True], indirect=True, ids=["loose", "strict"])

    @pytest.mark.skip()
    @parametrize
    async def test_method_namespaces(self, async_client: AsyncTurbopuffer) -> None:
        client = await async_client.namespaces()
        assert_matches_type(AsyncNamespacePage[NamespaceSummary], client, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_method_namespaces_with_all_params(self, async_client: AsyncTurbopuffer) -> None:
        client = await async_client.namespaces(
            cursor="cursor",
            page_size=1,
            prefix="prefix",
        )
        assert_matches_type(AsyncNamespacePage[NamespaceSummary], client, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_raw_response_namespaces(self, async_client: AsyncTurbopuffer) -> None:
        response = await async_client.with_raw_response.namespaces()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        client = await response.parse()
        assert_matches_type(AsyncNamespacePage[NamespaceSummary], client, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_streaming_response_namespaces(self, async_client: AsyncTurbopuffer) -> None:
        async with async_client.with_streaming_response.namespaces() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            client = await response.parse()
            assert_matches_type(AsyncNamespacePage[NamespaceSummary], client, path=["response"])

        assert cast(Any, response.is_closed) is True



import pytest

import turbopuffer
from turbopuffer import Turbopuffer, AsyncTurbopuffer
from tests.custom import test_prefix


def test_compression_auto_on(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "compression-auto-on")

    try:
        ns.delete_all()
    except turbopuffer.NotFoundError:
        pass

    ns.write(
        upsert_rows=[
            {
                "id": 1,
                "vector": [0.1] * 1024,
                "title": "test doc",
            },
        ],
        distance_metric="euclidean_squared",
    )

    # This request is large enough to be compressed.
    result = ns.query(
        rank_by=("vector", "ANN", [0.1] * 1024),
        top_k=1,
        include_attributes=True,
    )

    perf = result.performance
    assert perf.client_total_ms > 0
    assert perf.client_compress_ms > 0
    assert perf.client_response_ms is not None and perf.client_response_ms > 0
    assert perf.client_body_read_ms is not None and perf.client_body_read_ms > 0
    assert perf.client_deserialize_ms > 0


def test_compression_auto_off(tpuf: Turbopuffer):
    ns = tpuf.namespace(test_prefix + "compression-auto-off")

    try:
        ns.delete_all()
    except turbopuffer.NotFoundError:
        pass

    ns.write(
        upsert_rows=[
            {
                "id": 1,
                "vector": [0.1],
                "title": "test doc",
            },
        ],
        distance_metric="euclidean_squared",
    )

    # This request is too small to be compressed.
    result = ns.query(
        rank_by=("vector", "ANN", [0.1]),
        top_k=1,
        include_attributes=True,
    )

    perf = result.performance
    assert perf.client_total_ms > 0
    assert perf.client_compress_ms == 0
    assert perf.client_response_ms is not None and perf.client_response_ms > 0
    assert perf.client_body_read_ms is not None and perf.client_body_read_ms > 0
    assert perf.client_deserialize_ms > 0


@pytest.mark.asyncio
async def test_async_compression_auto_on(async_tpuf: AsyncTurbopuffer):
    ns = async_tpuf.namespace(test_prefix + "async-compression-auto-on")

    try:
        await ns.delete_all()
    except turbopuffer.NotFoundError:
        pass

    await ns.write(
        upsert_rows=[
            {
                "id": 1,
                "vector": [0.1] * 1024,
                "title": "test doc",
            },
        ],
        distance_metric="euclidean_squared",
    )

    # This request is large enough to be compressed.
    result = await ns.query(
        rank_by=("vector", "ANN", [0.1] * 1024),
        top_k=1,
        include_attributes=True,
    )

    perf = result.performance
    assert perf.client_total_ms > 0
    assert perf.client_compress_ms > 0
    assert perf.client_response_ms is not None and perf.client_response_ms > 0
    assert perf.client_body_read_ms is not None and perf.client_body_read_ms > 0
    assert perf.client_deserialize_ms > 0


async def test_async_compression_auto_off(async_tpuf: AsyncTurbopuffer):
    ns = async_tpuf.namespace(test_prefix + "async-compression-auto-off")

    try:
        await ns.delete_all()
    except turbopuffer.NotFoundError:
        pass

    await ns.write(
        upsert_rows=[
            {
                "id": 1,
                "vector": [0.1],
                "title": "test doc",
            },
        ],
        distance_metric="euclidean_squared",
    )

    # This request is too small to be compressed.
    result = await ns.query(
        rank_by=("vector", "ANN", [0.1]),
        top_k=1,
        include_attributes=True,
    )

    perf = result.performance
    assert perf.client_total_ms > 0
    assert perf.client_compress_ms == 0
    assert perf.client_response_ms is not None and perf.client_response_ms > 0
    assert perf.client_body_read_ms is not None and perf.client_body_read_ms > 0
    assert perf.client_deserialize_ms > 0



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from typing_extensions import Literal

from .._models import BaseModel

__all__ = ["NamespaceHintCacheWarmResponse"]


class NamespaceHintCacheWarmResponse(BaseModel):
    status: Literal["OK"]
    """The status of the request."""

    message: Optional[str] = None



import struct
from typing import List

import pybase64

from turbopuffer._utils._utils import lru_cache


@lru_cache(maxsize=64)
def _get_struct(vector_length: int) -> struct.Struct:
    return struct.Struct(f"<{vector_length}f")


def b64encode_vector(vector: List[float]) -> str:
    """
    Encodes a list of floats into a base64 string representation of the float
    array in a little-endian binary format.

    Args:
        vector (List[float]): The list of floats to encode.

    Returns:
        str: The base64 encoded string representation of the vector.
    """
    packed = _get_struct(len(vector)).pack(*vector)
    return pybase64.b64encode_as_string(packed)


def b64decode_vector(encoded: str) -> List[float]:
    """
    Decodes a base64 string representation of a float array in a little-endian
    binary format into a list of floats.

    Args:
        encoded (str): The base64 encoded string representation of the vector.

    Returns:
        List[float]: The decoded list of floats.
    """
    packed = pybase64.b64decode(encoded, validate=True)
    return list(_get_struct(len(packed) // 4).unpack(packed))



import time
from typing import Any, Dict, Optional

from .._models import BaseModel


class ClientPerformance(BaseModel):
    client_total_ms: float
    """Total request time in milliseconds, measured on the client.

    Note this only includes the time spent during the last retry.
    """

    client_compress_ms: float
    """Number of milliseconds spent compressing the request on the client."""

    client_response_ms: Optional[float]
    """Number of milliseconds between the start of the request and receiving the
    last response header from the server.
    """

    client_body_read_ms: Optional[float]
    """Number of milliseconds between receiving the last response header from
    the server and receiving the end of the response body from the server.
    """

    client_deserialize_ms: float
    """Number of milliseconds spent deserializing the response on the client."""


def merge_clock_into_response(response: Any, clock: Dict[str, float]) -> None:
    # HACK: we assume any response with a `performance` property of type
    # `ClientPerformance` wants client performance metrics merged into that
    # object.
    performance = getattr(response, "performance", None)
    if isinstance(performance, ClientPerformance):
        end = time.monotonic()

        # These always exist, regardless of transport.
        request_start: float = clock["request_start"]
        deserialize_start: float = clock["deserialize_start"]

        # These exist only when using our default (custom) transports.
        compress_start: Optional[float] = clock.get("compress_start")
        compress_end: Optional[float] = clock.get("compress_end")
        response_headers_end: Optional[float] = clock.get("response_headers_end")
        body_read_end: Optional[float] = clock.get("body_read_end")

        performance.client_total_ms = (end - request_start) * 1000
        performance.client_deserialize_ms = (end - deserialize_start) * 1000

        if compress_start and compress_end:
            performance.client_compress_ms = (compress_end - compress_start) * 1000

        if response_headers_end:
            performance.client_response_ms = (response_headers_end - request_start) * 1000

        if body_read_end and response_headers_end:
            performance.client_body_read_ms = (body_read_end - response_headers_end) * 1000



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, List, Union, Iterable
from typing_extensions import Required, TypeAlias, TypedDict

from .id_param import IDParam
from .vector_param import VectorParam

__all__ = ["ColumnsParam"]


class ColumnsParamTyped(TypedDict, total=False):
    id: Required[List[IDParam]]
    """The IDs of the documents."""

    vector: Union[List[VectorParam], Iterable[float], str]
    """The vector embeddings of the documents."""


ColumnsParam: TypeAlias = Union[ColumnsParamTyped, Dict[str, Iterable[object]]]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union
from typing_extensions import TypeAlias

from .full_text_search_config_param import FullTextSearchConfigParam

__all__ = ["FullTextSearchParam"]

FullTextSearchParam: TypeAlias = Union[bool, FullTextSearchConfigParam]



# Copyright (c) 2020 Florimond Manca
# Obtained from the following URL on 9 June 2025:
# https://gist.github.com/florimondmanca/d56764d78d748eb9f73165da388e546e
#
# In theory this should be faster than the default `httpx` transport, but in
# practice it seems to be slightly slower. We keep it here in case it's useful
# in the future.

import ssl
import gzip
import time
import typing
from typing_extensions import override

import httpx
import urllib3
from httpx._types import CertTypes, ProxyTypes, SyncByteStream
from httpx._config import create_ssl_context

from .transport import MIN_GZIP_SIZE


def httpx_headers_to_urllib3_headers(headers: httpx.Headers) -> urllib3.HTTPHeaderDict:
    urllib3_headers = urllib3.HTTPHeaderDict()
    for name, value in headers.multi_items():
        urllib3_headers.add(name, value)
    return urllib3_headers


class ResponseStream(SyncByteStream):
    CHUNK_SIZE = 1024

    def __init__(self, urllib3_stream: typing.Any) -> None:
        self._urllib3_stream = urllib3_stream

    @override
    def __iter__(self) -> typing.Iterator[bytes]:
        for chunk in self._urllib3_stream.stream(self.CHUNK_SIZE, decode_content=False):
            yield chunk

    @override
    def close(self) -> None:
        self._urllib3_stream.release_conn()


class Urllib3Transport(httpx.BaseTransport):
    def __init__(
        self,
        verify: typing.Union[ssl.SSLContext, str, bool] = True,
        trust_env: bool = True,
        max_pools: int = 10,
        maxsize: int = 10,
        cert: typing.Union[CertTypes, None] = None,
        proxy: typing.Union[ProxyTypes, None] = None,
    ) -> None:
        ssl_context = create_ssl_context(cert=cert, verify=verify, trust_env=trust_env)
        proxy = httpx.Proxy(url=proxy) if isinstance(proxy, (str, httpx.URL)) else proxy

        if proxy is None:
            self._pool = urllib3.PoolManager(
                ssl_context=ssl_context,
                num_pools=max_pools,
                maxsize=maxsize,
                block=False,
            )
        elif proxy.url.scheme in ("http", "https"):
            self._pool = urllib3.ProxyManager(
                str(proxy.url.netloc),
                num_pools=max_pools,
                maxsize=maxsize,
                block=False,
                proxy_ssl_context=proxy.ssl_context,
                proxy_headers=httpx_headers_to_urllib3_headers(proxy.headers),
                ssl_context=ssl_context,
            )
        elif proxy.url.scheme == "socks5":
            from urllib3.contrib.socks import SOCKSProxyManager

            username, password = proxy.auth or (None, None)

            self._pool = SOCKSProxyManager(
                proxy_url=str(proxy.url),
                num_pools=max_pools,
                maxsize=maxsize,
                block=False,
                username=username,
                password=password,
            )
        else:
            raise ValueError(
                f"Proxy protocol must be either 'http', 'https', or 'socks5', but got {proxy.url.scheme!r}."
            )

    @override
    def handle_request(self, request: httpx.Request) -> httpx.Response:
        clock = request.extensions["clock"]
        timeouts = request.extensions.get("timeout", {})

        connect_timeout = timeouts.get("connect", None)
        read_timeout = timeouts.get("read", None)

        urllib3_timeout = urllib3.Timeout(
            connect=connect_timeout,
            read=read_timeout,
        )

        # Compress the request content if worthwhile.
        content = request.content
        clock["compress_start"] = time.monotonic()
        clock["compress_end"] = clock["compress_start"]  # overwritten later if we actually compress
        if len(request.content) > MIN_GZIP_SIZE:
            content = gzip.compress(request.content, compresslevel=1)
            request.headers["Content-Encoding"] = "gzip"
            request.headers["Content-Length"] = str(len(content))
            clock["compress_end"] = time.monotonic()

        response = self._pool.request(
            request.method,
            str(request.url),
            body=content,
            headers=httpx_headers_to_urllib3_headers(request.headers),
            redirect=False,
            preload_content=False,
            timeout=urllib3_timeout,
        )
        clock["response_headers_end"] = time.monotonic()

        httpx_response = httpx.Response(
            status_code=response.status,
            headers=httpx.Headers([(name, value) for name, value in response.headers.iteritems()]),
            content=ResponseStream(response),
            extensions={
                # Smuggle the clock through the response object.
                "clock": clock,
            },
        )

        # Eagerly read the response content (which will be cached on the
        # response object) so that we can time the body read.
        httpx_response.read()
        clock["body_read_end"] = time.monotonic()

        return httpx_response



# Code generated by turbopuffer-apigen. DO NOT EDIT.

from typing import Any, Tuple, Union, Literal, Sequence

AggregateBy = Tuple[Literal["Count"], str]
Filter = Union[
    Tuple[str, Literal["Eq"], Any],
    Tuple[str, Literal["NotEq"], Any],
    Tuple[str, Literal["In"], Any],
    Tuple[str, Literal["NotIn"], Any],
    Tuple[str, Literal["Lt"], Any],
    Tuple[str, Literal["Lte"], Any],
    Tuple[str, Literal["Gt"], Any],
    Tuple[str, Literal["Gte"], Any],
    Tuple[str, Literal["Glob"], str],
    Tuple[str, Literal["NotGlob"], str],
    Tuple[str, Literal["IGlob"], str],
    Tuple[str, Literal["NotIGlob"], str],
    Tuple[str, Literal["ContainsAllTokens"], str],
    Tuple[str, Literal["ContainsAllTokens"], Sequence[str]],
    Tuple[Literal["Not"], "Filter"],
    Tuple[Literal["And"], Sequence["Filter"]],
    Tuple[Literal["Or"], Sequence["Filter"]],
]
RankByVector = Tuple[str, Literal["ANN"], Sequence[float]]
RankByText = Union[
    Tuple[str, Literal["BM25"], str],
    Tuple[str, Literal["BM25"], Sequence[str]],
    Tuple[Literal["Sum"], Sequence["RankByText"]],
    Tuple[Literal["Max"], Sequence["RankByText"]],
    Tuple[Literal["Product"], Tuple[float, "RankByText"]],
    Tuple[Literal["Product"], Tuple["RankByText", float]],
]
RankByAttributeOrder = Union[Literal["asc"], Literal["desc"]]
RankByAttribute = Tuple[str, RankByAttributeOrder]
RankBy = Union[RankByVector, RankByText, RankByAttribute]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .._models import BaseModel

__all__ = ["NamespaceSummary"]


class NamespaceSummary(BaseModel):
    id: str
    """The namespace ID."""
